/* $Id: ParametrizedTestSuite.java,v 1.1.1.1 2003/03/10 16:34:15 taweili Exp $ */

package de.tivano.junit;

import junit.framework.TestResult;
import junit.framework.AssertionFailedError;
import junit.framework.TestSuite;
import junit.framework.TestCase;
import junit.framework.Test;
import java.util.Iterator;
import java.util.Enumeration;
import java.util.TreeSet;
import java.util.Properties;
import java.util.Set;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Constructor;
import de.tivano.junit.ParametrizedTestSuite.Warning;
import de.tivano.junit.ParametrizedTestSuite.ParametrizedTestWrapper;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.io.InputStream;
import java.io.IOException;
import junit.framework.TestFailure;

/**
 * A unit test suite that executes the same test case for a number of
 * different sets of data.
 *
 * <p>Test data is defined as a standard Java {@link Properties} object, with a
 * certain convention on the property keys used: Keys are expected to be of
 * the form "&lt;test name&gt;.&lt;attribute name&gt;", where the test name part
 * may be any string allowed as part of a property key, and the attribute name
 * part must be a legal java identifier. The test name part is used to identify 
 * one of the test data definition in the property set (typically, a single 
 * property set contains many test data definitions), and the attribute name
 * is used to identify a single attribute within a test data definition.</p>
 *
 * <p>A test case is parametrized with a given test data definition by setting 
 * the accessible fields of this test case to the values specified by the 
 * corresponding property in the test data definition.  A field in the test
 * case is considered accessible if it either is a public field or if the
 * test case class defines a public setter method matching the attribute
 * name.</p> 
 *
 * <p>If both a public field and a setter method are present for a 
 * given attribute name, the setter method will be preferred. If more than one
 * matching setter method is found, the test will fail with an appropriate
 * error message.</p>
 * 
 * </ol>
 *
 * <p>Note that the fields that can be parametrized must have one of the 
 * types listed here:</p>
 * <ul>
 * <li>One of the primitive types <code>boolean</code>, <code>character</code>, 
 * <code>byte</code>, <code>short</code> <code>int</code>, <code>long</code>, 
 * <code>float</code> or <code>double</code>.</li>
 * <li>An object wrapper of a primitive type: {@link Boolean}, {@link Character},
 * {@link Byte}, {@link Short}, {@link Integer}, {@link Long}, {@link Float} or
 * {@link Double}</li>
 * <li>{@link String}</li>
 * <li>An arbitrary, non-abstract class that defines a public constructor taking
 * a single <code>String</code> as an argument (e.g. {@link java.io.File}, or
 * {@link java.net.URL}).
 * </ul>
 * <p>The same restrictions hold for the the parameters of setter methods. 
 * In addition, a setter method may only have a single parameter to be used 
 * for test case parametrization.</p>
 *
 * <p>Example: The property set
 * <pre>
 * Test1.count = 1
 * Test1.isSpecial = false;
 * Test1.url = http://www.tivano.de/
 *
 * Test2.count = 1
 * Test2.isSpecial = false;
 * Test2.url = file:///some/path
 *
 * My.Special.Test.count = 3
 * My.Special.Test.isSpecial = true
 * My.Special.Test.special = foo, bar, baz
 * </pre>
 * defines three test data sets named "Test1", "Test2" and "My.Special.Test". Test1 and
 * Test2 each define the attributes "count", "isSpecial" and "url", and My.Special.Test defines
 * the attributes "count", "isSpecial" and "special".</p> 
 * <p>Now, if the test class
 * <pre>
 * public class ExampleTest implements Test {
 *    public int count;
 *    public boolean isSpecial;
 *    private URL url;
 *    public void setIsSpecial(boolean value) { ... }
 *    public void setSpecial(String value) { ... }
 *    ....
 * }
 * </pre>
 * is parametrized by Test1, the variable <code>count</code> will be set to 1,
 * the method <code>setIsSpecial()</code> will be called with the argument
 * <code>false</code> and the variable <code>url</code> will be set to the equivalent of
 * <code>new URL("http://www.tivano.de/")</code>. The method 
 * <code>setSpecial()</code> will not be called, as there is no attribute named
 * "special" in Test1. When the example test is parametrized by My.Special.Test,
 * however, <code>setSpecial()</code> will be called with the argument 
 * "foo, bar, baz", and the variable <code>url</code> will be left alone.</p>
 * 
 * <p>Parametrization occurs after the test case constructor has finished
 * and before the test case is run. This means that you won't see the 
 * parametrized values in the constructor of your test cases, but they will be 
 * available in the <code>run(...)</code> and (for tests inheriting 
 * {@link TestCase}) <code>setUp()</code> methods.</p>
 *
 * <p>If a test case cannot be parametrized e.g. because there is no 
 * corresponding public field or setter method for an attribute, the
 * field or setter method violates the restrictions given above, or the
 * field or setter method is not public, the test will fail with an
 * appropriate error message.</p>
 * 
 * @author Richard Kunze
 */
public class ParametrizedTestSuite extends TestSuite {
    
    private Properties TEST_DATA = null;

    /** Helper class for reporting warnings. This test always fails. */
    static class Warning extends TestCase {
        private final String MESSAGE;
        public Warning(String message) {
            super("warning");
            MESSAGE = message;
        }
        protected void runTest() { fail(MESSAGE); }
    }
    
    /** Helper class for parametrizing a single test case.
     * The test is parametrized immediately before calling the
     * <code>run(TestResult)</code> method of the decorated test.
     */
    class ParametrizedTestWrapper extends TestCase {
        final Test WRAPPED;
        /** Construct a parametrized wrapper for <code>test</code>. 
         * @param key the key prefix for the associated test data set
         */
        public ParametrizedTestWrapper(Test test) {
            super(null);
            if (test instanceof TestCase)
                setName(((TestCase)test).getName());        
            WRAPPED = test;            
        }
        
        public Test getTest() { return WRAPPED; }
        
        /** Parametrize and run the test */
        public void run(TestResult result) {
            result.startTest(this);
            if (parametrizeTest(result)) {                
                // Collect the results in a temporary test result. Any errors or
                // failures are added to the result later, substituting this
                // object for the wrapped test.
                TestResult tmp = new TestResult();
                WRAPPED.run(tmp);
                Enumeration errors = tmp.errors();
                while (errors.hasMoreElements()) {
                    Throwable err = ((TestFailure)errors.nextElement()).thrownException();
                    result.addError(this, err);
                }
                Enumeration failures = tmp.failures();
                while (failures.hasMoreElements()) {
                    AssertionFailedError err = 
                       (AssertionFailedError)((TestFailure)failures.nextElement()).thrownException();
                    result.addFailure(this, err);
                }                                       
            }
            result.endTest(this);
        }
               
        /** Get the data set key prefix for this test. */
        private String getKeyPrefix() {
            return ParametrizedTestSuite.this.getName();
        }
        
        /** Parametrize the decorated test. If there are errors in parametrizing the
         * test, appropriate failure messages are added to <code>result</code> and 
         * this method returns <code>false</code>.
         */
        private boolean parametrizeTest(TestResult result) {
            boolean success = true;
            Enumeration keys = TEST_DATA.keys();
            String keyPrefix = getKeyPrefix() + ".";
            int prefixLen = keyPrefix.length();
            Test test = WRAPPED;
            Class testClass = test.getClass();
            while (keys.hasMoreElements()) {
                String key = keys.nextElement().toString();
                // Skip this entry if the key doesn't belong to the data
                // set for this test case, or if it is too short to contain
                // an attribute name.
                if (!key.startsWith(keyPrefix) || key.length() <= prefixLen) {
                    continue;
                }
                String attributeName = key.substring(prefixLen);
                // Find the approriate field/setter method, and determine the
                // field/argument type.
                try {
                    Method setterMethod = 
                        findSetterMethod(testClass, attributeName);
                    Field field = null;
                    if (setterMethod == null) 
                        field = findPublicField(testClass, attributeName);
                    if (setterMethod == null && field == null) {
                        fail("Cannot set value for attribute '" + attributeName + 
                              "': No setter method or public member field found " + 
                              "with that name.");
                    }
                    Object value;
                    if (setterMethod != null) {
                        // Getting the first parameter cannot fail here because
                        // findSetterMethod() already checks that the method has
                        // exactly one parameter.
                        value = parseAttributeValue(attributeName,
                                                    TEST_DATA.getProperty(key),
                                                    setterMethod.getParameterTypes()[0]);
                        
                        Object[] args = { value };
                        try {
                            setterMethod.invoke(test, args);
                        } catch (InvocationTargetException e) {
                            Throwable cause = e.getTargetException();
                            if (cause instanceof RuntimeException) {
                                throw (RuntimeException)cause;
                            } else if (cause instanceof Error) {
                                throw (Error)cause;
                            } else fail("Exception in " + setterMethod.getName() + 
                                        ": " + cause.toString());
                        } catch (IllegalAccessException e) {
                            fail("Cannot set value for attribute '" + attributeName +
                                 "': " + e.toString());
                        }

                    } else {
                        value = parseAttributeValue(attributeName,
                                                    TEST_DATA.getProperty(key),
                                                    field.getType());
                        try {
                            field.set(test, value);
                        } catch (IllegalAccessException e) {
                            fail("Cannot set value for attribute '" + attributeName +
                                 "': " + e.toString());
                        }
                    }
                } catch (AssertionFailedError e) {
                    result.addFailure(this, e);
                    success = false;
                }
            }
            return success;
        }
        
        /** Find the appropriate setter method for <code>attrName</code>.
         * If more than one setter method is found or if no matching setter
         * is found at all, this method throws an <code>AssertionFailedError</code>
         * with appropriate error message.
         * @param testClass the test class
         * @param attrName the name of the attribute to find a setter method for.
         */
        private Method findSetterMethod(Class testClass, String attrName) {
            String methodName = "set" + Character.toUpperCase(attrName.charAt(0))
                                + attrName.substring(1);
            Method[] allMethods = testClass.getMethods();
            
            // Find a suitable method. At the same time, count all methods 
            // that are suitable candidates, and check if there are methods
            // with a matching name at all.
            Method candidate = null;
            boolean foundMatchingName = false;
            int matchingMethodCount = 0;
            for (int i=0; i<allMethods.length; i++) {
                if (!allMethods[i].getName().equals(methodName)) continue;
                foundMatchingName = true;
                Class[] params = allMethods[i].getParameterTypes();
                if (params.length != 1) continue;
                if (isValidAttributeType(params[0])) {
                    candidate = allMethods[i];
                    matchingMethodCount++;
                }
            }
            // If there is no method at all with a matching name, return
            // null so that the caller goes on checking for a public field.
            if (!foundMatchingName) return null;
            
            // OK, we have a matching name. Now check if there is exactly one
            // method with this name that has a matching signature (i.e., exactly
            // one argument of a valid type). If this is not hte case, report an
            // appropriate error.
            if (matchingMethodCount != 1) {
                fail("Cannot set value for attribute '" + attrName + "': " 
                     + (matchingMethodCount==0?"No":"More than one")
                     + " setter method with compatible parameter types found.");
            }
            
            // Still here? fine, we've found our setter...
            return candidate;
        }
        
        /** Helper method to find a public field of matching type for 
         * <code>fieldName</code>. If no field with that name is found, this 
         * method returns <code>null</code>. If a field is found but does not
         * have a valid type, this method throws an <code>AssertionFailedError</code>
         * with appropriate error message.
         */
        private Field findPublicField(Class testClass, String fieldName) {
            try {
                Field field = testClass.getField(fieldName);
                if (!isValidAttributeType(field.getType())) {
                    fail("Cannot set value for attribute '" + fieldName + 
                         "': Public member field has incompatible type '" +
                         field.getType().getName() + "'");
                }
                return field;
            } catch (NoSuchFieldException e) {
                return null;
            }
        }
        
        /** Check if <code>type</code> is a type compatible with parametrization. */
        private boolean isValidAttributeType(Class type) {
            if (type.isPrimitive() && !type.equals(Void.TYPE)) return true;
            if (Boolean.class.equals(type) || Character.class.equals(type)
                || Byte.class.equals(type) || Short.class.equals(type)
                || Integer.class.equals(type) || Long.class.equals(type)
                || Float.class.equals(type) || Double.class.equals(type)
                || String.class.equals(type)) return true;
            
            // Not a primitve, an object wrapper of a primitive, or a string ->
            // Check if type has a public constructor taking a string as an
            // argument.
            final Class[] ARG_TYPES = { String.class };
            try {
                type.getConstructor(ARG_TYPES);
                return true;
            } catch (NoSuchMethodException e) {
                return false;
            }
        }
        
        /** Helper method to parse a string into an object of the desired type.
         * For primitive types, this method returns the appropriate object wrapper.
         * @throws AssertionFailedError if the string cannot be parsed.
         */
        private Object parseAttributeValue(String name, String value, Class type) {
            try {
                if (value == null) return value;
                if (Boolean.TYPE.equals(type) || Boolean.class.equals(type)) {
                    return Boolean.valueOf(value);
                } else if (Character.TYPE.equals(type) || Character.class.equals(type)) {
                    if (value.length() != 1) {
                        fail("Cannot set attribute '" + name + "' to value '" +
                             value + "': Value too long.");
                    }
                    return new Character(value.charAt(0));
                } else if (Byte.TYPE.equals(type) || Byte.class.equals(type)) {
                    return Byte.valueOf(value);
                } else if (Short.TYPE.equals(type) || Short.class.equals(type)) {
                    return Short.valueOf(value);
                } else if (Integer.TYPE.equals(type) || Integer.class.equals(type)) {
                    return Integer.valueOf(value);
                } else if (Long.TYPE.equals(type) || Long.class.equals(type)) {
                    return Long.valueOf(value);
                } else if (Float.TYPE.equals(type) || Float.class.equals(type)) {
                    return Float.valueOf(value);
                } else if (Double.TYPE.equals(type) || Double.class.equals(type)) {
                    return Double.valueOf(value);
                } else if (String.class.equals(type)) {
                    return value;
                } else {
                    final Class[] ARG_TYPES = { String.class };
                    Constructor c = type.getConstructor(ARG_TYPES);
                    try {
                        final Object[] ARG = { value };
                        return c.newInstance(ARG);
                    } catch (InvocationTargetException e) {
                        throw e.getTargetException();
                    }
                }
            } catch (AssertionFailedError e) { 
                throw e;
            } catch (Throwable t) {
                fail("Cannot set attribute '" + name + "' to value '" +
                      value + "': " + t.toString());
                // this is only here to keep the compiler happy - fail()
                // always throws an exception...
                return null;
            }
        }
        
        public String toString() { 
            return WRAPPED.toString() + "#" + getKeyPrefix();
        }
        
    }
        
    /** Helper method to prepare the test data. This method always returns
     * a <code>Properties</code> object, even if the argument is <code>null</code>.
     * In this case, and when the argument is empty, a warning with an appropriate
     * error message is added to this test suite.
     * @param rawData the "raw" test data. May be <code>null</code>
     * @param clone if <code>true</code>, the test data is cloned.
     * @return a <code>Properties</code> object containing the test data.
     */
    private Properties prepareTestData(Properties rawData, boolean clone) {
        if (rawData == null) {
            super.addTest(warning("No test data specified for test suite '" + getName() + "'!"));
            return new Properties();
        } else if (rawData.isEmpty()) {
            super.addTest(warning("Test data definition for test suite '" + getName() + "' is empty!"));
        }        
        if (clone) return (Properties)rawData.clone();
        else return rawData;
    }
    
    /** 
     * Helper method to find the test data for <code>name</code>. This method tries 
     * to look up a resource named "<em>name</em>.properties" and parse this resource
     * into a properties object. If this succeeds, the resulting properties object
     * is returned. Otherwise, this method adds a warning with an appropriate message
     * to the test suite and returns an empty properties object.
     * @param name the name of the test data resource
     */
    private Properties findTestData(String name) {
        String resourceName = name + ".properties";
        InputStream res = getClass().getResourceAsStream(resourceName);
        
        Properties prop = new Properties();
        if (res == null) {
            super.addTest(warning("Cannot load test data: Resource '" 
                                  + resourceName + "' not found!"));
        } else {
            try {
                prop.load(res);
            } catch (IOException e) {
                super.addTest(warning("Cannot load test data from resource '" 
                              + resourceName +"': " + e.toString()));
            }
        }
        return prepareTestData(prop, false);
    }
    
    /** 
     * Helper method to find the test data for <code>testClass</code>. This method tries 
     * to look up a resource with a name corresponding to the name of the test class with
     * all occurences of "." replaced by "/", a prefix of "/" and a suffix of ".propertes".
     * If this succeeds, the resulting properties object is returned. Otherwise, this method 
     * adds a warning with an appropriate message to the test suite and returns an empty 
     * properties object.
     * @param testClass the class to find test data for
     */
    private Properties findTestData(Class testClass) {
        return findTestData("/" + testClass.getName().replace('.', '/'));
    }
    
    /** Construct an empty, parametrized test suite for the given test data set.
     * All tests that are subsequently added to this suite using the 
     * {@link #addTest} or {@link #addTestSuite} methods will be parametrized
     * according to <code>testData</code>.
     * @param testData the property set defining the test data.
     */
    public ParametrizedTestSuite(Properties testData) {
        this(null, testData);
    }
    
    /** Construct an empty test suite named <code>name</code> for the
     * test data set <code>testData</code>.
     * All tests that are subsequently added to this suite using the 
     * {@link #addTest} or {@link #addTestSuite} methods will be parametrized
     * according to <code>testData</code>.
     * @param the name for this test suite.
     * @param testData the property set defining the test data.
     */
    public ParametrizedTestSuite(String name, Properties testData) {
        super(name);
        TEST_DATA = prepareTestData(testData, true);
    }
    
    /** Construct an empty, parametrized test suite named <code>name</code>.
     * <p>The test data for this test suite is expected to reside in a resource
     * named "<em>name</em>.properties". If no such resource exists, the resource
     * cannot be parsed into a properties object, or the corresponding properties
     * object is empty, a warning with an appropriate error message is added
     * to this test suite.</p>
     * @param name the name for this test suite
     */
    public ParametrizedTestSuite(String name) {
        super(name);
        TEST_DATA = findTestData(name);
    }
    
    /** Construct a parametrized test suite for a given test class.
     * <p>The test data for this suite is expected to reside in a resource
     * with a name matching that of the test class (in detail: The resource 
     * name is constructed by replacing all occurrences of "." in the class name
     * by "/" and appending the suffix ".properties"), and it will contain tests
     * for all test methods defined by <code>testClass</code> and all test data sets
     * specified in the associated test data resource.</p> 
     *
     * <p>Using this constructor is the exact equivalent of the code sequence
     * <pre>
     * String name = "/" + testClass.getName().replace(".", "/") + ".properties"
     * TestSuite suite = new ParametrizedTestSuite(name);
     * suite.addTestSuite(testClass);
     * </pre>
     * </p>
     * 
     * <p>See the documentation for {@link #addTestSuite} for an exact definition
     * of the tests that are constructed by using this constructor.</p>
     *
     * <p>If the resource defining the test data set for the test class cannot 
     * be found, cannot be parsed as a property set or if the resulting property
     * set is empty, a warning will be added to this suite instead of the
     * actual test. Likewise, a warning will be added if one of the resulting
     * tests cannot be parametrized (e.g because of a mismatch of the test data
     * and the fields respectively setter methods defined by the test class).</p>
     * @param testClass the test class.
     */
    public ParametrizedTestSuite(Class testClass) {
        super(testClass.getName());
        TEST_DATA = findTestData(testClass);
        if (!TEST_DATA.isEmpty()) {
            addTestSuite(testClass);
        }
    }
    
    /** 
     * Special constructor for {@link #addTestSuite}. The constructed test suite
     * will hold tests for all test methods in <code>testClass</code>, parametrized
     * according to the test data set named <code>name</code> defined in
     * <code>testData</code>. This constructor assumes that <code>testData</code>
     * is not null or empty and that <code>name</code> is a valid key prefix for
     * a test data set defined in <code>testData</code>.
     */
    private ParametrizedTestSuite(Class testClass, String name, Properties testData) {
        // Implementation note: The current implementation assumes that the 
        // TestSuite(Class) constructor calls addTest() for every test it
        // extracts from testClass. Since addTest() wraps its argument in an
        // ParametrizedTestWrapper if this argument is not a warning, 
        // simply calling super(testClass) is enough to construct the complete suite.
        // We need to re-set the name, though, because the name is used by the
        // ParametrizedTestWrapper to look up the appropriate data set, and super(class)
        // sets the name to be testClass.getName().
        super(testClass);
        TEST_DATA = testData;
        setName(name);
    }
    
    /** Returns a test that will fail with a warning message. 
     * This method must be redefined here for two reasons: First,
     * it is private in {@link TestSuite} and so not accessible here. 
     * Second, the implementation in <code>TestSuite</code> uses an
     * anonymous class for the warning, which makes it impossible for
     * {@link #addTest} to distinguish between warnings that should not
     * be wrapped in a {@link de.tivano.junit.ParametrizedTestSuite.ParametrizedTestWrapper} 
     * (because doing so will almost certainly cause the wrapper to fail with 
     * a number of spurious errors when it tries to parametrize the warning) 
     * and real tests, which have to be wrapped.
     * @param message the warning message
     */
    protected Test warning(String message) {
        return new Warning(message);
    }
    
    /** Get the set of key prefixes for all test data sets defined in this test suite.
     * @return a {@link Set} of {@link Strings}
     */
    private Set getTestDataKeys() {
        // We're using a tree set here to make sure the test case keys are
        // sorted by name.
        Set keys = new TreeSet();
        Enumeration allKeys = TEST_DATA.keys();
        while (allKeys.hasMoreElements()) {
            String rawKey = allKeys.nextElement().toString();
            int endIdx = rawKey.lastIndexOf('.');
            keys.add((endIdx!=-1?rawKey.substring(0,endIdx):rawKey));
        }
        return keys;
    }
    
    /** Add the parametrized test suite for <code>testClass</code> to this suite.
     * 
     * <p>For every different test data set found in the associated property set, 
     * a test suite will be created and added to this suite that contains tests 
     * for all <code>test...()</code> method found in the test class. Every one of 
     * these tests is parametrized by the corresponding test data set.</p>
     *
     * <p>Example: If the class
     * <pre>
     * public class ExampleTest extends TestCase {
     *    private String value;
     *    public void setValue(String str) { ... }
     *    public void testFoo() { ... }
     *    public void testBar() { ... }
     * }
     * </pre>
     * is passed as parameter <code>testClass</code> to this method, and the property 
     * set associated with this test suite has the content
     * <pre>
     * Test1.value = fnord
     * Test2.value = snafu
     * </pre>
     * then this method will create two test suites named "Test1" and "Test2" 
     * and add them to this suite. Each of the created test suites contains 
     * two instances of ExampleTest: One that runs the <code>testFoo()</code> 
     * test method, and one that runs <code>testBar()</code>. The tests in the 
     * "Test1" suite will have the <code>value</code> attribute set to "fnord", 
     * the tests in the "Test2" suite to "snafu".</p>
     * @param testClass the test class
     */
    public void addTestSuite(Class testClass) {
        Iterator keys = getTestDataKeys().iterator();
        while (keys.hasNext()) {
            // Need to use super.addTest() to prevent addTest() from wrapping
            // the newly created test suite.
            super.addTest(new ParametrizedTestSuite(testClass, 
                                                    keys.next().toString(), 
                                                    TEST_DATA));
        }
    }

    /** Adds a test to the suite.
     * <p>If a test data set with a key of <code>this.getName()</code> exists in the test
     * data associated with this test suite, the test will be parametrized with the
     * test data in this data set.</p>
     * <p>Note that if <code>test</code> is an instance of {@link TestSuite}, then it will <em>not</em> 
     * parametrized.</p>
     * @param test the test to be added to this suite.
     */
    public void addTest(Test test) {
        // Don't try and parametrize warnings - doing so will almost certainly
        // fail and produce spurious error reports.
        // Don't parametrize test suites - this doesn't really make sense anyway,
        // and the interjected test wrapper gets in the way of tools that examine
        // the test hierarchy (e.g. the junit.swingui.TestRunner).
        if (test instanceof Warning || test instanceof TestSuite) super.addTest(test);
        else super.addTest(new ParametrizedTestWrapper(test));
    }
}
