/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DOMInfo.java,v 1.1.1.1 2003/03/10 16:36:12 taweili Exp $
 */

package tests.org.enhydra.xml.xhtml.dominfo;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringReader;
import java.util.Map;
import java.util.TreeMap;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import java.lang.reflect.Method;
import com.lutris.util.FatalExceptionError;

/**
 * Class to collection information about the objects that compose an
 * set of DOM interfaces or implementations.
 */
class DOMInfo {
    /*
     * A property of an Element class.  This is defined by get and set methods.
     */
    public class ElementProperty {
        /** The property name */
        private String fName;

        /** Lower-case version of the name */
        private String fNameLower;

        /** Accessor */
        private Method fAccessor; 

        /** Mutator */
        private Method fMutator;
        
        /** Construct a new property */
        ElementProperty(String name) {
            fName = name;
            fNameLower = name.toLowerCase();
        }

        /**
         * Add an accessor/mutator method.
         */
        void addAccessor(Method method) {
            if (method.getName().startsWith("get")) {
                fAccessor = method;
            } else {
                fMutator = method;
            }
        }

        /** Get the property name */
        public String getName() {
            return fName;
        }

        /** Get the lower-case property name */
        public String getNameLower() {
            return fNameLower;
        }

        /** Get the attribute name the property represents. */
        public String getAttrName() {
            return fNameLower;
        }

        /** Get the accessor method */
        public Method getAccessor() {
            return fAccessor;
        }
        
        /** get the mutator method */
        public Method getMutator() {
            return fMutator;
        }

        /* Dump the class */
        public void dump(PrintWriter out,
                         int level) {
            PrintUtils.printIndent(out, level);
            out.println(fName + ": " + 
                        ((fAccessor == null) ? "null" : fAccessor.getName())
                        + "/" + 
                        ((fMutator == null) ? "null" : fMutator.getName()));
        }
    }

    /**
     * Information about a class or interface.
     */
    public class ElementClass {
        /** The Class object for the Element class or interface */
        private Class fClassObj;

        /**
         * Interfaces this implements.  If this class is an interface,
         * then this is the interface it extends.
         */
        private Map fImplements = new TreeMap();
 
        /**
         * List of properties.  A property is someting with a get* and/or
         * set* method.  This includes methods that are inherrited.  Key
         * is the lower-case name.
         */
        private Map fProperties = new TreeMap();

        /** Process a get/set method */
        private void processPropertyMethod(Method method) {
            String name = method.getName().substring(3);
            String nameLower = name.toLowerCase();
            ElementProperty prop = (ElementProperty)fProperties.get(nameLower);
            if (prop == null) {
                prop = new ElementProperty(name);
                fProperties.put(prop.getNameLower(), prop);
            }
            prop.addAccessor(method);
        }
        
        /** Record information about a method */
        private void processMethod(Method method) {
            if (isPropertyMethod(method)) {
                processPropertyMethod(method);
            }
        }

        /** Find the properties */
        private void findProperties() {
            Method[] methods = fClassObj.getMethods();
            for (int idx = 0; idx < methods.length; idx++) {
                processMethod(methods[idx]);
            }
        }

        /**
         * Process a class to determine if it should be added to
         * the implements list.  Recursively search its interfaces and
         * super class.
         */
        private void processImplements(Class classObj) {
            if (classObj.isInterface() && !fImplements.containsKey(classObj.getName())) {
                fImplements.put(classObj.getName(), classObj);
                findImplements(classObj);
            }
        }

        /**
         * Construct a list of implemented interfaces, walking the
         * inheritance heierarchy to find all interfaces.
         */
        private void findImplements(Class classObj) {
            Class[] ifaces = classObj.getInterfaces();
            for (int idx = 0; idx < ifaces.length; idx++) {
                if (!fImplements.containsKey(ifaces[idx].getName())) {
                    fImplements.put(ifaces[idx].getName(), ifaces[idx]);
                    findImplements(ifaces[idx]);
                }
            }
            Class superObj = classObj.getSuperclass();
            if (superObj != null) {
                processImplements(superObj);
            }
        }

        /** Constructor */
        public ElementClass(Class classObj) {
            fClassObj = classObj;
            findProperties();
            findImplements(fClassObj);
        }

        /** Get the class name */
        public String getName() {
            return fClassObj.getName();
        }

        /* Get the lower-case property names */
        public Set getPropertyNames() {
            return fProperties.keySet();
        }

        /* Get an property */
        public ElementProperty getProperty(String name) {
            ElementProperty prop = (ElementProperty)fProperties.get(name.toLowerCase());
            if (prop == null) {
                throw new Error("Can't find property \"" + name + "\" of element \"" + getName() + "\"");
            }
            return prop;
        }

        /* Get the implemented interface names */
        public Set getImplementsNames() {
            return fImplements.keySet();
        }

        /* Dump the class */
        public void dump(PrintWriter out,
                         int level) {
            PrintUtils.printIndent(out, level);
            out.println(getName());

            PrintUtils.printIndent(out, level+1);
            out.println("Implements:");
            Iterator impls = getImplementsNames().iterator();
            while (impls.hasNext()) {
                PrintUtils.printIndent(out, level+2);
                out.println((String)impls.next());
            }

            PrintUtils.printIndent(out, level+1);
            out.println("Properties:");
            Iterator propNames = getPropertyNames().iterator();
            while (propNames.hasNext()) {
                getProperty((String)propNames.next()).dump(out, level+2);
            }
        }

        /** Get a string representation */
        public String toString() {
            return fClassObj.getName();
        }
    }

    /** Element names to drop (lower-cased) */
    private Set fDropElementNames = new HashSet();

    /** Methods to ignore when getting properties. */
    private Set fDropPropertyMethodNames = new HashSet();

    /** The elements in the DOM */
    private Map fElements = new TreeMap();

    /** Should an element be dropped */
    private boolean dropElementClass(String className) {
        return fDropElementNames.contains(className.toLowerCase());
    }

    /** Is this a property method (get* or set* and not in drop list) */
    private boolean isPropertyMethod(Method method) {
        String name = method.getName();
        return ((name.startsWith("get") || name.startsWith("set")) && !fDropPropertyMethodNames.contains(name));
    } 

    /** Add method to the drop property list */
    private void addDropPropertyMethod(String name) {
        fDropPropertyMethodNames.add(name);
    }
    
    /** Add the property-like methods in a class to a list of ones to ignore */
    private void addDropPropertyMethods(Class classObj) {
        Method[] methods = classObj.getMethods();
        for (int idx = 0; idx < methods.length; idx++) {
            if (isPropertyMethod(methods[idx])) {
                fDropPropertyMethodNames.add(methods[idx].getName());
            }
        }
    }

    /* Constructor */
    public DOMInfo() {
        addDropPropertyMethods(org.w3c.dom.Element.class);
        addDropPropertyMethods(org.enhydra.apache.xerces.dom.ElementImpl.class);
        addDropPropertyMethods(org.enhydra.apache.xerces.dom.ElementNSImpl.class);
    }


    /* Constructor with package */
    public DOMInfo(String pkgName) {
        this();
        addElementsFromClassPath(pkgName);
    }

    /* Get the element names */
    public Set getElementNames() {
        return fElements.keySet();
    }

    /* Get an element */
    public ElementClass getElement(String name) {
        ElementClass element = (ElementClass)fElements.get(name);
        if (element == null) {
            throw new Error("Can't find element \"" + name + "\"");
        }
        return element;
    }

    /* Get the set of ElementClass objects */
    public Set getElementClassSet() {
        HashSet set = new HashSet();

        Iterator elementNames = getElementNames().iterator();
        while (elementNames.hasNext()) {
            set.add(getElement((String)elementNames.next()));
        }
        return set;
    }

    /* Add an element */
    private void addElement(ElementClass element) {
        if (fElements.containsKey(element.getName())) {
            throw new Error("Element class info already exists:  " + element.getName());
        }
        fElements.put(element.getName(), element);
    }

    /** Add an element given its class */
    public void addElement(Class element) {
        addElement(new ElementClass(element));
    }

    /** Drop a suffix from a name */
    private String dropSuffix(String suffix,
                              String name) {
        if (name.regionMatches(name.length()-suffix.length(), suffix, 0, suffix.length())) {
            name = name.substring(0, name.length()-suffix.length());
        }
        return name;
    }

    /** Clean up a unqualifiied class name (making unqualified if necessary) */
    private String cleanUnqualClassName(String name) {
        // Delete .class and .java suffixs, if they exist
        name = dropSuffix(".class", name);
        name = dropSuffix(".java", name);

        // Delete package qualification
        int idx = name.lastIndexOf('.');
        if (idx >= 0) {
            name = name.substring(idx+1);
        }
        return name;
    }
    

    /** Load an element class and add */
    public void addElement(String className) {
        try {
            addElement(Class.forName(className));
        } catch (ClassNotFoundException except) {
            throw new FatalExceptionError(except);
        }
    }

    /** Load an element class and add */
    public void addElement(String pkgName,
                           String unqualClassName) {
        String className = pkgName + "." + cleanUnqualClassName(unqualClassName);
        try {
            addElement(Class.forName(className));
        } catch (ClassNotFoundException except) {
            throw new FatalExceptionError(except);
        }
    }

    /** Load a list of class in a package */
    public void addElements(String pkgName,
                            String[] unqualClassNames) {
        for (int idx = 0; idx < unqualClassNames.length; idx++) {
            addElement(pkgName, unqualClassNames[idx]);
        }
    }

    /**
     * Load a list of class in a package by searching the CLASSPATH
     * for the specified package.
     */
    public void addElementsFromClassPath(String pkgName) {
        String[] classes = ElementClassMappings.getPackageElementClasses(pkgName);
        for (int idx = 0; idx < classes.length; idx++) {
            addElement(classes[idx]);
        }
    }

    /**
     * Dump the elements in the DOM.
     */
    public void dump(PrintWriter out) {
          Iterator elementNames = getElementNames().iterator();
          while (elementNames.hasNext()) {
              getElement((String)elementNames.next()).dump(out, 0);
          }
    }

    /**
     * Main used for testing.
     */
    public static void main(String[] args) throws IOException {
        if (args.length < 1) {
            System.err.println("Wrong # args: DOMInfo package [classFile1 classFile2 ...]]");
            System.exit(1);
        }
        DOMInfo domInfo = new DOMInfo();

        if (args.length == 1) {
            domInfo.addElementsFromClassPath(args[0]);
        } else {
            for (int idx = 1; idx < args.length; idx++) {
                domInfo.addElement(args[0], args[idx]);
            }
        }
        domInfo.dump(new PrintWriter(System.out, true));
    }
}
