/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: Relations.java,v 1.1.1.1 2003/03/10 16:36:12 taweili Exp $
 */

package tests.org.enhydra.xml.xhtml.dominfo;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringReader;
import java.util.Map;
import java.util.TreeMap;
import java.util.Set;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;
import java.lang.reflect.Method;

/**
 * Build a table mapping the relationship between the XHTML DOM interfaces,
 * implementations and the W3C HTML interfaces.
 */
class Relations {
    /* Indices used hwne addressing the set of DOM classes */
    public static final int HTML_INTERFACE_IDX = 0;
    public static final int XHTML_INTERFACE_IDX = 1;
    public static final int XHTML_IMPLEMENTATION_IDX = 2;
    public static final int NUM_DOM_CLASS_TYPES = 3;

    /**
     * A tag relation to DOM classes.
     */
    public class TagRelation {
        /* Tag name */
        private final String fTagName;
        
        /* DTD information */
        private final DTDInfo.ElementInfo fElementInfo;

        /* Element class information for each DOM. */
        private DOMInfo.ElementClass[] fElementClassInfo = new DOMInfo.ElementClass[NUM_DOM_CLASS_TYPES];

        /*
         * Attribute relations for the tag, keyed by lower-case attribute name.
         * Each value is a linked-list of AttrRelations.
         */
        private Map fAttrRelations = new TreeMap();

        /** Constructor */
        public TagRelation(DTDInfo.ElementInfo elementInfo) {
            fTagName = elementInfo.getRawName();
            fElementInfo = elementInfo;

            // Create AttrRelation objects.
            Iterator attrNames = elementInfo.getAttrNames().iterator();
            while (attrNames.hasNext()) {
                addDTDAttrInfo(elementInfo.getAttr((String)attrNames.next()));
            }
        }

        /** Constructor without elementInfo (for special tag mappings) */
        public TagRelation(String tagName) {
            fTagName = tagName;
            fElementInfo = null;
        }
        
        /** Add a new DTDInfo.AttrInfo object. */
        private void addDTDAttrInfo(DTDInfo.AttrInfo attrInfo) {
            // NB name is lower-cased by AttrRelation
            AttrRelation attrRelation = new AttrRelation(this, attrInfo);
            if (fAttrRelations.containsKey(attrRelation.getAttrName())) {
                throw new Error("attribute \"" + attrRelation.getAttrName() + "\" already exists");
            }
            fAttrRelations.put(attrRelation.getAttrName(), attrRelation);
        }

        
        /** Add the properties to the AttrRelation for the tag */
        private void addElementProperties(int domIdx,
                                          DOMInfo.ElementClass elementClassInfo) {
            Iterator props = elementClassInfo.getPropertyNames().iterator();
            while (props.hasNext()) {
                addElementProperty(domIdx, elementClassInfo.getProperty((String)props.next()));
            }
        }

        /** Get the names for the AttributeRelations. */
        public Set getAttrRelationNames() {
            return fAttrRelations.keySet();
        }

        /** An attribute relation (which maybe chained) */
        public AttrRelation getAttrRelation(String name) {
            return (AttrRelation)fAttrRelations.get(name);
        }

        /** Get the tag name */
        public String getTagName() {
            return fTagName;
        }

        /** Get the DTDInfo.ElementInfo object. */
        public DTDInfo.ElementInfo getElementInfo() {
            return fElementInfo;
        }

        /** Get the element clas info for a DOM set */
        public DOMInfo.ElementClass getElementClassInfo(int domIdx) {
            return fElementClassInfo[domIdx];
        }

        /** Set the element class info for a DOM set */
        public void setElementClassInfo(int domIdx,
                                        DOMInfo.ElementClass elementClassInfo) {
            fElementClassInfo[domIdx] = elementClassInfo;
            //FIXME:            System.err.println("setElementClassInfo" + domIdx);
            addElementProperties(domIdx, elementClassInfo);
        }

        /** Add a property to the AttrRelation for the tag */
        private void addElementProperty(int domIdx,
                                        DOMInfo.ElementProperty prop) {
            String attrName = prop.getAttrName();
            if (fSpecialPropertyAttrMap.containsKey(attrName)) {
                attrName = (String)fSpecialPropertyAttrMap.get(attrName);
            }

            AttrRelation attrRelHead = (AttrRelation)fAttrRelations.get(attrName);
            if (attrRelHead == null) {
                // Hmm, no entry for this in DTD!
                attrRelHead = new AttrRelation(this, prop.getAttrName());
                fAttrRelations.put(attrRelHead.getAttrName(), attrRelHead);
            }

            // Find the first one with an empty entry for this DOM idx.
            AttrRelation attrRel = attrRelHead;
            while ((attrRel != null) && (attrRel.getElementPropertyInfo(domIdx) != null)) {
                attrRel = attrRel.getNextAttrRelation();
            }
            if (attrRel == null) {
                // If available, construct with attrInfo 
                DTDInfo.AttrInfo attrInfo = attrRelHead.getAttrInfo();
                if (attrInfo != null) {
                    attrRel = new AttrRelation(this, attrInfo);
                } else {
                    attrRel = new AttrRelation(this, prop.getAttrName());
                }
                attrRelHead.addAttrRelation(attrRel);
            }
            attrRel.setElementPropertyInfo(domIdx, prop);
        }
    }

    /** 
     * A relationship between an attribute of a tag and the properties
     * detected on the DOM classes.
     */
    public class AttrRelation {
        /* Tag associated with */
        private final TagRelation fTagRelation;

        /* Attribute name (lower-cased) */
        private final String fAttrName;
        
        /* DTD information */
        private final DTDInfo.AttrInfo fAttrInfo;

        /* Element class information for each DOM. */
        private DOMInfo.ElementProperty[] fElementPropertyInfo = new DOMInfo.ElementProperty[NUM_DOM_CLASS_TYPES];

        /**
         * If multiple properties map to the same method, due to case
         * insensitive compares, different access method mapping to the same
         * attribute, etc.
         */
        private AttrRelation fNext;

        /** Constructor */
        public AttrRelation(TagRelation tagRelation,
                            DTDInfo.AttrInfo attrInfo) {
            fTagRelation = tagRelation;
            fAttrName = attrInfo.getRawName().toLowerCase();
            fAttrInfo = attrInfo;
            //FIXME:            System.err.println("AttrRelation: " + tagRelation.getTagName() + ": " + fAttrName);
        }

        /** Constructor with no DTDInfo.AttrInfo */
        public AttrRelation(TagRelation tagRelation,
                            String attrNameLower) {
            fTagRelation = tagRelation;
            fAttrName = attrNameLower;
            fAttrInfo = null;
            //FIXME: System.err.println("AttrRelation: " + tagRelation.getTagName() + ": " + fAttrName);
        }

        /* Get tag associated with */
        public TagRelation  getTagRelation() {
            return fTagRelation;
        }

        /** Get the tag name */
        public String getAttrName() {
            return fAttrName;
        }

        /** Get the DTDInfo.AttrInfo object. */
        public DTDInfo.AttrInfo getAttrInfo() {
            return fAttrInfo;
        }

        /** Get the attr clas info for a DOM set */
        public DOMInfo.ElementProperty getElementPropertyInfo(int domIdx) {
            return fElementPropertyInfo[domIdx];
        }

        /** Set the element clas info for a DOM set */
        public void setElementPropertyInfo(int domIdx,
                                           DOMInfo.ElementProperty elementPropertyInfo) {
            fElementPropertyInfo[domIdx] = elementPropertyInfo;
        }

        /**
         * Add a next AttrRelation.  This will traverse to the end of the chain and
         * add at the tail.
         */
        public void addAttrRelation(AttrRelation next) {
            if (fNext == null) {
                fNext = next;
            } else {
                fNext.addAttrRelation(next);
            }
        }

        /**
         * Get next AttrRelation in the chain.
         */
        public AttrRelation getNextAttrRelation() {
            return fNext;
        }
    }


    /** Parsed DTD */
    private DTDInfo fDTDInfo;

    /** DOMInfo objects */
    private DOMInfo[] fDOMInfos = new DOMInfo[NUM_DOM_CLASS_TYPES];

    /**
     * DOMInfo.ElementClass objects that have not been added to the relation
     * table.
     */
    private Set[] fPendingElementClasses = new Set[NUM_DOM_CLASS_TYPES];

    /** Tag relations tables, indexed by tag name. */
    private Map fTags = new TreeMap();

    /** Special mappings of property names to attribute names (global for now); mapped to lowercase */
    private static final String[][] SPECIAL_PROPERTY_ATTR_MAP = {
        {"ClassName", "class"},
        {"ChOff", "charoff"},
        {"Ch", "char"}
    };
    private static final HashMap fSpecialPropertyAttrMap = new HashMap();

    static {
        for (int idx = 0; idx < SPECIAL_PROPERTY_ATTR_MAP.length; idx++) {
            fSpecialPropertyAttrMap.put(SPECIAL_PROPERTY_ATTR_MAP[idx][0].toLowerCase(),
                                        SPECIAL_PROPERTY_ATTR_MAP[idx][1]);
        }
    }

    /**
     * Construct the DOM tables starting with the given DTD.
     */
    public Relations(String dtdPath) {
        fDTDInfo = new DTDInfo(dtdPath);
        buildDOMInfos();
        buildTagsTable();
    }

    /** Get the DTDInfo object */
    public DTDInfo getDTDInfo() {
        return fDTDInfo;
    }

    /*
     * Initialize the object for a particular DOM class set.
     */
    private void buildDOMInfo(int domIdx,
                              String domPackage) {
        fDOMInfos[domIdx] = new DOMInfo(domPackage);
        fPendingElementClasses[domIdx] = fDOMInfos[domIdx].getElementClassSet();
    }
    
    /**
     * Create the DOM info objects.
     */
    private void buildDOMInfos() {
        buildDOMInfo(HTML_INTERFACE_IDX, "org.w3c.dom.html");
        buildDOMInfo(XHTML_INTERFACE_IDX, "org.enhydra.xml.xhtml.dom");
        buildDOMInfo(XHTML_IMPLEMENTATION_IDX, "org.enhydra.xml.xhtml.dom.xerces");
    }

    /**
     * Get the Element class for a domIdx
     */
    private Class getDOMElement(String tag,
                                int domIdx) {
        switch (domIdx) {
        case HTML_INTERFACE_IDX:
            return ElementClassMappings.getHTMLInterface(tag);
        case XHTML_INTERFACE_IDX:
            return ElementClassMappings.getXHTMLInterface(tag);
        case XHTML_IMPLEMENTATION_IDX:
            return ElementClassMappings.getXHTMLImplementation(tag);
        default:
            return null;
        }
    }

    /**
     * Link a DOMInfo.ElementClass object to the tag.
     */
    private void addTagElementClass(TagRelation tagRel,
                                    int domIdx) {
        Class elementClass = getDOMElement(tagRel.getTagName(), domIdx);
        DOMInfo.ElementClass elementClassInfo = fDOMInfos[domIdx].getElement(elementClass.getName());
        tagRel.setElementClassInfo(domIdx, elementClassInfo);
        fPendingElementClasses[domIdx].remove(elementClassInfo);
    }

    /**
     * Add a tag to that table.
     */
    private void addTag(String tagName) {
        TagRelation tagRel = new TagRelation(fDTDInfo.getElement(tagName));
        fTags.put(tagName, tagRel);
        for (int domIdx = 0; domIdx < NUM_DOM_CLASS_TYPES; domIdx++) {
            addTagElementClass(tagRel, domIdx);
        }
    }

    /**
     * Add a special tag mapping if it doesn't exist.
     */
    private void addSpecialTag(String elementClassName,
                               String tagName,
                               int domIdx) {
        TagRelation tagRel = (TagRelation)fTags.get(tagName);
        if (tagRel == null) {
            tagRel = new TagRelation(tagName);
            fTags.put(tagName, tagRel);
        }
        if (tagRel.getElementClassInfo(domIdx) == null) {
            DOMInfo.ElementClass elementClassInfo = fDOMInfos[domIdx].getElement(elementClassName);
            tagRel.setElementClassInfo(domIdx, elementClassInfo);
            fPendingElementClasses[domIdx].remove(elementClassInfo);
        }
    }

    /**
     * Build the tags table.
     */
    private void buildTagsTable() {
        Iterator elementNames = fDTDInfo.getElementNames().iterator();
        while (elementNames.hasNext()) {
            addTag((String)elementNames.next());
        }

        // Add in special elements/tag mappings if they are not already there
        addSpecialTag(org.enhydra.xml.xhtml.dom.XHTMLElement.class.getName(), "[element]", XHTML_INTERFACE_IDX);
        addSpecialTag(org.enhydra.xml.xhtml.dom.xerces.XHTMLElementImpl.class.getName(), "[element]", XHTML_IMPLEMENTATION_IDX);
        addSpecialTag(org.w3c.dom.html.HTMLElement.class.getName(), "[element]", HTML_INTERFACE_IDX);

        addSpecialTag(org.enhydra.xml.xhtml.dom.XHTMLFrameSetElement.class.getName(), "frameset", XHTML_INTERFACE_IDX);
        addSpecialTag(org.enhydra.xml.xhtml.dom.xerces.XHTMLFrameSetElementImpl.class.getName(), "frameset",XHTML_IMPLEMENTATION_IDX);
    }

    /* Get the tag names */
    public Set getTagNames() {
        return fTags.keySet();
    }

    /* Get a TagRelation */
    public TagRelation getTagRelation(String tagName) {
        TagRelation tagRel = (TagRelation)fTags.get(tagName);
        if (tagRel == null) {
            throw new Error("Can't find tag \"" + tagName + "\"");
        }
        return tagRel;
    }

    /*
     * Get the set of DOMInfo.ElementClass objects not associated
     * with a tag.
     */
    public Set getUnmappedElementClasses(int domIdx) {
        return fPendingElementClasses[domIdx];
    }

    /**
     * Main used for testing.
     */
    public static void main(String[] args) {
        if (args.length != 1) {
            System.err.println("Wrong # args: Relations dtdfile");
            System.exit(1);
        }
        Relations relations = new Relations(args[0]);
    }
}
