/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: Report.java,v 1.1.1.1 2003/03/10 16:36:12 taweili Exp $
 */

package tests.org.enhydra.xml.xhtml.dominfo;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringReader;
import java.util.Map;
import java.util.TreeMap;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import java.lang.reflect.Method;

/**
 * Generate a report on the state on the DOM relations.
 * This gotta kind of kludge, needs redo if it has a long life.
 */
public class Report {
    /* Number of columns */
    private static final int NUM_COLUMNS = 4;

    /** Column assignments  */
    private static final int TAG_COLUMN_IDX = 0;
    private static final int HTML_INTER_COLUMN = 1;
    private static final int XHTML_INTER_COLUMN = 2;
    private static final int XHTML_IMPL_COLUMN = 3;

    /** Column widths */
    private static final int[] COLUMN_WIDTHS = {
        12, 25, 25, 25
    };

    /** Output file being written */
    private PrintWriter fOut;

    /** Column buffers */
    private StringBuffer[] fColumns = new StringBuffer[NUM_COLUMNS];

    /** Lines to divide up output */
    private String fDoubleLine;
    private String fSingleLine;

    /** Object describing the DOM relationships */
    private Relations fRelations;

    /** Should `clean' attribute relations non be printed */
    private boolean fSkipCleanAttrRelations;

    /**
     * Constructor.
     */
    public Report(String dtdPath,
                  PrintWriter out,
                  boolean skipCleanAttrRelations) {
        for (int col = 0; col < NUM_COLUMNS; col++) {
            fColumns[col] = new StringBuffer(4096);
        }
        fRelations = new Relations(dtdPath);
        fOut = out;
        fDoubleLine = makeLine('=');
        fSingleLine = makeLine('-');
        fSkipCleanAttrRelations = skipCleanAttrRelations;
    }

    /**
     * Create a line the width of the output.
     */
    private String makeLine(char ch) {
        int len = 0;
        for (int idx = 0; idx < COLUMN_WIDTHS.length; idx++) {
            len += COLUMN_WIDTHS[idx];
        }
        StringBuffer buf = new StringBuffer(len);
        for (int cnt = 0; cnt < len; cnt++) {
            buf.append(ch);
        }
        return buf.toString();
    }

    /** Convert a name to an unqualified name */
    private String unqualName(String name) {
        int idx = name.lastIndexOf('.');
        if (idx < 0) {
            return name;
        } else {
            return name.substring(idx+1);
        }
    }

    /** Append a string to the current line in a column */
    private void print(int col,
                       String s) {
        fColumns[col].append(s);
    }

    /** Add a new line to a column */
    private void println(int col) {
        fColumns[col].append('\n');
    }

    /** Write and string and flush current line */
    private void println(int col,
                         String s) {
        print(col, s);
        println(col);
    }

    /** Print a line in a column */
    private void printSingleLine(int col) {
        fColumns[col].append(fSingleLine.substring(0, COLUMN_WIDTHS[col]));
        println(col);
    }
    
    /** Check if the next char of all columns is past the end */
    private boolean allDone(int[] nextChar,
                            String[] columns) {
        for (int col = 0; col < NUM_COLUMNS; col++) {
            if (nextChar[col] < columns[col].length()) {
                return false;
            }
        }
        return true;
    }

    /** Find the numer of chars upto a newline */
    private int charsOnLine(String column,
                            int nextChar) {
        if (nextChar >= column.length()) {
            return 0;
        } else {
            int nlIdx = column.indexOf('\n', nextChar);
            if (nlIdx < 0) {
                return (column.length() - nextChar);
            } else {
                return (nlIdx - nextChar);
            }
        }
    }

    /** Write a column, return update nextChar */
    private int writeCol(String column,
                         int nextChar,
                         int width) {
        int charCnt = charsOnLine(column, nextChar);
        int writeCnt = (charCnt < width) ? charCnt : width;
        if (writeCnt > 0) {
            fOut.write(column, nextChar, writeCnt);
        }
        for (int cnt = charCnt; cnt < width; cnt++) {
            fOut.write(' ');
        }
        return nextChar + charCnt + 1;
    }

    /** Write the columns */
    private void flush() {
        int[] nextChar = new int[NUM_COLUMNS];
        String[] columns = new String[NUM_COLUMNS];
        for (int col = 0; col < NUM_COLUMNS; col++) {
            columns[col] = fColumns[col].toString();
            fColumns[col].setLength(0);
        }
        // Write column groups
        while (!allDone(nextChar, columns)) {
            // Write a line
            for (int col = 0; col < NUM_COLUMNS; col++) {
                nextChar[col] = writeCol(columns[col], nextChar[col],
                                         COLUMN_WIDTHS[col]);
            }
            fOut.println();
        }
    }

    /** Output DTDInfo.ElementInfo name column */
    private void writeElementInfoName(Relations.TagRelation tagRel,
                                      int col) {
        DTDInfo.ElementInfo elementInfo = tagRel.getElementInfo();
        if (elementInfo != null) {
            println(col, elementInfo.getRawName());
        } else {
            println(col, tagRel.getTagName());
        }
    }

    /** Output DTDInfo.ElementInfo attributes column */
    private void writeElementInfoAttrs(DTDInfo.ElementInfo elementInfo,
                                       int col) {
        printSingleLine(col);
        Iterator attrNames = elementInfo.getAttrNames().iterator();
        while (attrNames.hasNext()) {
            println(col, (String)attrNames.next());
        }
    }

    /** Write the DOMInfo.ElementClass name */
    private void writeElementClassName(DOMInfo.ElementClass elementClass,
                                       int col) {
        println(col, unqualName(elementClass.getName()));
    }

    /** Write DOMInfo.ElementClass implements */
    private void writeElementClassImplements(DOMInfo.ElementClass elementClass,
                                             int col) {
        printSingleLine(col);
        Iterator impls = elementClass.getImplementsNames().iterator();
        while (impls.hasNext()) {
            String name = (String)impls.next();
            if (!(name.equals(org.w3c.dom.Node.class.getName())
                  || name.equals(org.w3c.dom.Element.class.getName()))) {
                println(col, unqualName(name));
            }
        }
    }

    /** Write DOMInfo.ElementClass properties */
    private void writeElementClassProps(DOMInfo.ElementClass elementClass,
                                        int col) {
        printSingleLine(col);
        Iterator propNames = elementClass.getPropertyNames().iterator();
        while (propNames.hasNext()) {
            DOMInfo.ElementProperty prop = elementClass.getProperty((String)propNames.next());
            println(col, prop.getName());
        }
    }

    /**
     * Write a DOMInfo.ElementProperty save in an AttrRelation.
     */
    private void writeAttrRelationProp(DOMInfo.ElementProperty elementPropertyInfo,
                                       int col) {
        if (elementPropertyInfo == null) {
            println(col, "*");
        } else {
            println(col, elementPropertyInfo.getName());
        }
    }

    /** Get the element property name for checking */
    private String getPropName(Relations.AttrRelation attrRel,
                               int domIdx) {
        DOMInfo.ElementProperty elementPropertyInfo = attrRel.getElementPropertyInfo(domIdx);
        if (elementPropertyInfo == null) {
            return null;
        } else {
            return elementPropertyInfo.getName();
        }
    }

    /**
     * Attributes and properties special-cases to always skip when skiping clean.
     */
    private static Set fAttrIsClean = new HashSet();
    private static Set fXHTMLPropIsClean = new HashSet();
    static {
        fAttrIsClean.add("onblur");
        fAttrIsClean.add("onclick");
        fAttrIsClean.add("ondblclick");
        fAttrIsClean.add("onfocus");
        fAttrIsClean.add("onkeydown");
        fAttrIsClean.add("onkeypress");
        fAttrIsClean.add("onkeyup");
        fAttrIsClean.add("onmousedown");
        fAttrIsClean.add("onmousemove");
        fAttrIsClean.add("onmouseout");
        fAttrIsClean.add("onmouseover");
        fAttrIsClean.add("onmouseup");
        fAttrIsClean.add("style");
        fAttrIsClean.add("xml:lang");
        fXHTMLPropIsClean.add("XmlLang");
    }

    /** Determine if an attrRel should be skipped. */
    private boolean skipAttrRelation(Relations.AttrRelation attrRel,
                                     boolean first) {
        if (fSkipCleanAttrRelations) {
            if ((attrRel.getNextAttrRelation() != null) || !first) {
                // have chained entries
                return false;
            }
            // Check property names
            String htmlName = getPropName(attrRel, Relations.HTML_INTERFACE_IDX);
            String xhtmlIFaceName = getPropName(attrRel, Relations.XHTML_INTERFACE_IDX);
            String xhtmlImplName = getPropName(attrRel, Relations.XHTML_IMPLEMENTATION_IDX);


            // Some known special cases.
            if (fAttrIsClean.contains(attrRel.getAttrName())
                || ((xhtmlIFaceName != null) && fXHTMLPropIsClean.contains(xhtmlIFaceName))
                || ((xhtmlImplName != null) && fXHTMLPropIsClean.contains(xhtmlImplName))) {
                return true; // special exemption
            }

            if (attrRel.getAttrInfo() == null) {
                // Don't actually have an attribute
                return false;
            }

            if ((htmlName != null) && htmlName.equals(xhtmlIFaceName) && htmlName.equals(xhtmlImplName)) {
                return true; // all match
            }
        }
        return false;
    }
    

    /**
     * Write the AttrRelation object.  This doesn't follow the chain.
     */
    private void writeAttrRelation(Relations.AttrRelation attrRel,
                                   boolean first) {
        if (skipAttrRelation(attrRel, first)) {
            return;
        }

        // Only print name on first in chain and if an attribute was actually
        // found; don't use assumed name
        if (first) {
            if (attrRel.getAttrInfo() != null) {
                println(TAG_COLUMN_IDX, attrRel.getAttrName());
            } else {
                println(TAG_COLUMN_IDX, "*");
            }
        }
        writeAttrRelationProp(attrRel.getElementPropertyInfo(Relations.HTML_INTERFACE_IDX),
                              HTML_INTER_COLUMN);
        writeAttrRelationProp(attrRel.getElementPropertyInfo(Relations.XHTML_INTERFACE_IDX),
                              XHTML_INTER_COLUMN);
        writeAttrRelationProp(attrRel.getElementPropertyInfo(Relations.XHTML_IMPLEMENTATION_IDX),
                              XHTML_IMPL_COLUMN);
    }

    /**
     * Write a AttrRelation chain.
     */
    private void writeAttrRelationChain(Relations.AttrRelation attrRel) {
        boolean first = true;
        while (attrRel != null) {
            writeAttrRelation(attrRel, first);
            first = false;
            attrRel = attrRel.getNextAttrRelation();
        }
        flush();
    }

    /**
     * Write the attribute relations for an element.
     */
    private void writeAttrRelations(Relations.TagRelation tagRel) {
        for (int col = 0; col < NUM_COLUMNS; col++) {
            printSingleLine(col);
        }

        Iterator attrNames = tagRel.getAttrRelationNames().iterator();
        while (attrNames.hasNext()) {
            writeAttrRelationChain(tagRel.getAttrRelation((String)attrNames.next()));
        }
    }

    /** Generate a report for a tag */
    private void tagReport(Relations.TagRelation tagRel) {
        fOut.println(fDoubleLine);
        writeElementInfoName(tagRel, TAG_COLUMN_IDX);
        writeElementClassName(tagRel.getElementClassInfo(Relations.HTML_INTERFACE_IDX),
                              HTML_INTER_COLUMN);
        writeElementClassName(tagRel.getElementClassInfo(Relations.XHTML_INTERFACE_IDX),
                              XHTML_INTER_COLUMN);
        writeElementClassName(tagRel.getElementClassInfo(Relations.XHTML_IMPLEMENTATION_IDX),
                              XHTML_IMPL_COLUMN);
        flush();
        printSingleLine(TAG_COLUMN_IDX);
        writeElementClassImplements(tagRel.getElementClassInfo(Relations.HTML_INTERFACE_IDX),
                                    HTML_INTER_COLUMN);
        writeElementClassImplements(tagRel.getElementClassInfo(Relations.XHTML_INTERFACE_IDX),
                                    XHTML_INTER_COLUMN);
        writeElementClassImplements(tagRel.getElementClassInfo(Relations.XHTML_IMPLEMENTATION_IDX),
                                    XHTML_IMPL_COLUMN);
        flush();
        writeAttrRelations(tagRel);
    }


    /** Process a DOMInfo class that didn't map to a tag */
    private void processUnusedElementClass(DOMInfo.ElementClass elementClass,
                                           int col) {
        fOut.println(fDoubleLine);
        println(TAG_COLUMN_IDX, "*UNUSED*");
        writeElementClassName(elementClass, col);
        writeElementClassImplements(elementClass, col);
        writeElementClassProps(elementClass, col);
        flush();
    }
    
    /** Generate unused element classes column */
    private void unusedElementClassColumn(Set unusedElementClasses,
                                          int col) {
        Iterator elementClasses = unusedElementClasses.iterator();
        while (elementClasses.hasNext()) {
            processUnusedElementClass((DOMInfo.ElementClass)elementClasses.next(), col);
        }
    }

    /** Generate unused element classes report */
    private void unusedElementClassReport() {
        unusedElementClassColumn(fRelations.getUnmappedElementClasses(Relations.HTML_INTERFACE_IDX),
                                 HTML_INTER_COLUMN);
        unusedElementClassColumn(fRelations.getUnmappedElementClasses(Relations.XHTML_INTERFACE_IDX),
                                 XHTML_INTER_COLUMN);
        unusedElementClassColumn(fRelations.getUnmappedElementClasses(Relations.XHTML_IMPLEMENTATION_IDX),
                                 XHTML_IMPL_COLUMN);
    }

    /** Generate a report */
    public void report() {
        fOut.println("DTD: " + fRelations.getDTDInfo().getDTDPath());
        Iterator tagNames = fRelations.getTagNames().iterator();
        while (tagNames.hasNext()) {
            tagReport(fRelations.getTagRelation((String)tagNames.next()));
        }
        unusedElementClassReport();
        fOut.flush();
    }

    /**
     * Main used for testing.
     */
    public static void main(String[] args) {
        int argIdx = 0;
        boolean skipCleanAttrRelations = false;
        while ((argIdx < args.length) && args[argIdx].startsWith("-")) {
            if (args[argIdx].equals("-skip-clean-attr")) {
                skipCleanAttrRelations = true;
            } else {
                System.err.println("bogus option: " + args[argIdx]);
            }
            argIdx++;
        }
        if ((args.length - argIdx) != 1) {
            System.err.println("Wrong # args: Report [-skip-clean-attr] dtdfile");
            System.exit(1);
        }
        Report report = new Report(args[argIdx], new PrintWriter(System.out),
                                   skipCleanAttrRelations);
        report.report();
    }
}
