/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: LazyDocumentType.java,v 1.2 2003/06/29 05:58:48 jkjome Exp $
 */

package org.enhydra.xml.lazydom;

import org.w3c.dom.*;
import org.enhydra.apache.xerces.dom.DocumentTypeImpl;
import org.enhydra.apache.xerces.dom.NodeImpl;

/**
 * Implementation of the DOM Attr that supports lazy instantiation of
 * a template DOM.
 */
public class LazyDocumentType extends DocumentTypeImpl implements LazyNode {
    /** 
     * Constructor.
     * @param ownerDoc The document that owns this node.
     * @param template If not-null, get the parameters from this template.
     * @param name Document type name.
     *  Will be ignored if template is not-null.
     * @param publicId The public id.
     *  Will be ignored if template is not-null.
     * @param systemId The system id.
     *  Will be ignored if template is not-null.
     * @param internalSubset A string representation of the internal subset.
     *  Will be ignored if template is not-null.
     */
    public LazyDocumentType(LazyDocument ownerDoc,
                            LazyDocumentType template,
                            String name, 
                            String publicId,
                            String systemId,
                            String internalSubset) {
    	super(ownerDoc, 
              (template != null) ? template.getNodeName() : name, 
              (template != null) ? template.getPublicId() : publicId,
              (template != null) ? template.getSystemId() : systemId);
        if (template != null) {
            fTemplateNode = template;
            fNodeId = template.getNodeId();
            setInternalSubset(template.getInternalSubset());
        } else {
            // No template, mark as expanded.
            fContentsExpanded = true;
            setInternalSubset(internalSubset);
        }
    }

    /**
     * Set the owner of the document.  Use to solve chicken and egg
     * problem with creating document with document type.
     */
    protected void setOwnerDocument(LazyDocument doc) {
        super.setOwnerDocument(doc);
    }

    //-------------------------------------------------------------------------
    // LazyDocumentType specific
    //-------------------------------------------------------------------------

    /**
     * Template for this DocumentType.
     */
    private LazyDocumentType fTemplateNode = null;

    /**
     * Have the contained nodes been expanded
     */
    private boolean fContentsExpanded = false;

    /**
     * Get the template for this node.
     * @see LazyNode#getTemplateNode
     */
    public LazyDocumentType getTemplateDocumentType() {
        return fTemplateNode;
    }

    /**
     * @see Node#cloneNode
     */
    public Node cloneNode(boolean deep) {
        if (deep) {
            // If children are copied, we must expand now.
            if (!fContentsExpanded) {
                expandContents();
            }
        }
        
        // This does a clone(), must clean up all fields.
        LazyDocumentType newDocumentType = (LazyDocumentType)super.cloneNode(deep);
        newDocumentType.fNodeId = NULL_NODE_ID;
        newDocumentType.fContentsExpanded = true;
        return newDocumentType;
    }

    /** 
     * Is the contents expanded?
     */
    public boolean isContentsExpanded() {
        return fContentsExpanded;
    }

    /**
     * Do work of expanding the contents.  Must be called synchronized
     * on the document.
     */
    private void doExpandContents(LazyDocument doc) {
        NamedNodeMap templateEntities = fTemplateNode.getEntities();
        int numEntities = templateEntities.getLength();
        NamedNodeMap entities = super.getEntities();
        for (int idx = 0; idx < numEntities; idx++) {
            entities.setNamedItem(((LazyNode)templateEntities.item(idx)).templateClone(doc));
        }

        NamedNodeMap templateNotations = fTemplateNode.getNotations();
        int numNotations = templateNotations.getLength();
        NamedNodeMap notations = super.getNotations();
        for (int idx = 0; idx < numNotations; idx++) {
            notations.setNamedItem(((LazyNode)templateNotations.item(idx)).templateClone(doc));
        }

        fContentsExpanded = true;
    }

    /*
     * Expand the contents.
     */
    private void expandContents() {
        LazyDocument doc = (LazyDocument)getOwnerDocument();
        synchronized (doc) {
            if (fContentsExpanded) {
                doExpandContents(doc);
            }
        }
    }

    /**
     * @see org.w3c.dom.DocumentType#getEntities
     */
    public NamedNodeMap getEntities() {
        if (!fContentsExpanded) {
            expandContents();
        }
        return super.getEntities();
    }

    /**
     * @see org.w3c.dom.DocumentType#getNotations
     */
    public NamedNodeMap getNotations() {
        if (!fContentsExpanded) {
            expandContents();
        }
        return super.getNotations();
    }

    //-------------------------------------------------------------------------
    // LazyNode support
    //-------------------------------------------------------------------------

    /*
     * Node id for this element.
     */
    private int fNodeId = NULL_NODE_ID;

    /**
     * Is this a template node?
     */
    private boolean fIsTemplateNode;

    /*
     * @see LazyNode#makeTemplateNode
     */
    public void makeTemplateNode(int nodeId) {
        fNodeId = nodeId;
        fIsTemplateNode = true;
    }

    /**
     * @see LazyNode#getNodeId
     */
    public int getNodeId() {
        return fNodeId;
    }

    /**
     * @see LazyNode#isTemplateNode
     */
    public boolean isTemplateNode() {
        return fIsTemplateNode;
    }

    /**
     * @see LazyNode#getTemplateNode
     */
    public LazyNode getTemplateNode() {
        return fTemplateNode;
    }

    /**
     * @see LazyNode#templateClone
     */
    public LazyNode templateClone(Document ownerDocument) {
        return new LazyDocumentType((LazyDocument)ownerDocument, this,
                                    null, null, null, null);
    }

    /**
     * Set the node value, invalidating the id.  All node data is modified
     * by this routine.
     * @see org.w3c.dom.Node#setNodeValue
     */
    public void setNodeValue(String value) {
        fNodeId = NULL_NODE_ID;
        super.setNodeValue(value);
    }

}
