/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: NodeCreateGenerator.java,v 1.1.1.1 2003/03/10 16:36:22 taweili Exp $
 */

package org.enhydra.xml.xmlc.dom.generic;

import org.enhydra.xml.xmlc.XMLCError;
import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.dom.XMLCDocument;
import org.enhydra.xml.xmlc.codegen.JavaLang;
import org.enhydra.xml.xmlc.codegen.JavaCode;
import org.w3c.dom.*;
import org.enhydra.xml.xmlc.compiler.ClassGenerator;

/**
 * Class used create DOM nodes.  This maybe extended by a class needing to
 * create nodes in a different.  This is called by the document builder.  The
 * code to generate adds is also in here, as added certain type of data to the
 * DOM is not specified in level 2.
 */
public class NodeCreateGenerator {
    /** Document object being compiled */
    private Document fDocument;

    /** The document element. */
    private Element fDocElement;

    /** Error messages */
    private static final String ENTITIES_UNSUPPORTED
        = "creating entity nodes not implemented for this DOM";
    private static final String NOTATION_UNSUPPORTED
        = "creating notation nodes not implemented for this DOM";
    
    /**
     * Constructor.
     */
    public NodeCreateGenerator(XMLCDocument xmlcDoc) {
        fDocument = xmlcDoc.getDocument();
        fDocElement = fDocument.getDocumentElement();
    }

    /** Is this the document element? */
    protected boolean isDocElement(Element element) {
        return (element == fDocElement);
    }

    /**
     * Generate Document and DocumentType creation. 
     */
    protected void genDocumentCreate(Document document,
                                     String docVar,
                                     String docTypeVar,
                                     JavaCode body) {
        DocumentType docType = document.getDoctype();
        String docTypeRef;
        if (docType == null) {
            docTypeRef = "null";
        } else {
            docTypeRef = docTypeVar;
            genDocumentTypeCreate(docType, docVar, docTypeVar, body);
        }

        String docClass = document.getClass().getName();
        body.addln(docClass + " " + docVar + " = (" + docClass + ")"
                   + ClassGenerator.DOM_FACTORY_FIELD_NAME + ".createDocument("
                   + JavaLang.createStringConst(fDocElement.getNamespaceURI()) + ", "
                   + JavaLang.createStringConst(fDocElement.getNodeName()) + ", "
                   + docTypeRef + ");");
    }
    
    /**
     * Generate DocumentType creation.
     */
    protected void genDocumentTypeCreate(DocumentType docType,
                                         String docVar,
                                         String varName, 
                                         JavaCode body) {
        String docTypeClass = docType.getClass().getName();
        body.addln(docTypeClass + " " + varName
                   + " = (" + docTypeClass + ")" 
                   + ClassGenerator.DOM_FACTORY_FIELD_NAME
                   + ".createDocumentType("
                   + JavaLang.createStringConst(docType.getNodeName()) + ", "
                   + JavaLang.createStringConst(docType.getPublicId()) + ", " 
                   + JavaLang.createStringConst(docType.getSystemId()) + ", " 
                   + JavaLang.createStringConst(docType.getInternalSubset()) + ");");
    }

    /**
     * Generate Entity creation.
     */
    protected void genEntityCreate(Entity entity,
                                   String docVar,
                                   String varName, 
                                   JavaCode body) {
        throw new XMLCError(ENTITIES_UNSUPPORTED);
    }

    /**
     * Generate Attr creation.
     */
    protected void genAttrCreate(Attr attr,
                                 String docVar,
                                 String varName, 
                                 JavaCode body) {
        body.add(varName + " = ");
        if (attr.getNamespaceURI() == null) {
            body.addln(docVar + ".createAttribute("
                       + JavaLang.createStringConst(attr.getName()) + ");");
        } else {
            body.addln(docVar + ".createAttributeNS("
                       + JavaLang.createStringConst(attr.getNamespaceURI()) + ", "
                       + JavaLang.createStringConst(attr.getName()) + ");");
        }
    }

    /**
     * Generate CDATASection creation.
     */
    protected void genCDATASectionCreate(CDATASection cdataSection,
                                         String docVar,
                                         String varName, 
                                         JavaCode body) {
        body.addln(varName + " = " + docVar + ".createCDATASection("
                   + JavaLang.createStringConst(cdataSection.getData()) + ");");
    }

    /**
     * Generate Comment creation.
     */
    protected void genCommentCreate(Comment comment,
                                    String docVar,
                                    String varName, 
                                    JavaCode body) {
        body.addln(varName + " = " + docVar + ".createComment("
                   + JavaLang.createStringConst(comment.getData()) + ");");
    }

    /**
     * Generate Element creation.
     */
    protected void genElementCreate(Element element,
                                    String docVar,
                                    String varName, 
                                    JavaCode body) {
        body.add(varName + " = ");
        if (isDocElement(element)) {
            // document element is created when document is created, just
            // get it
            body.addln(docVar + ".getDocumentElement();");
        } if (element.getNamespaceURI() == null) {
            body.addln(docVar + ".createElement("
                       + JavaLang.createStringConst(element.getTagName()) + ");");
        } else {
            body.addln(docVar + ".createElementNS("
                       + JavaLang.createStringConst(element.getNamespaceURI()) + ", "
                       + JavaLang.createStringConst(element.getTagName()) + ");");
        }
    }

    /**
     * Generate EntityReference creation.
     */
    protected void genEntityReferenceCreate(EntityReference entityRef,
                                            String docVar,
                                            String varName, 
                                            JavaCode body) {
        body.addln(varName + " = " + docVar
                   + ".createEntityReference("
                   + JavaLang.createStringConst(entityRef.getNodeName())
                   + ");");
    }

    /**
     * Generate Notation creation.
     */
    protected void genNotationCreate(Notation notation,
                                     String docVar,
                                     String varName, 
                                     JavaCode body) {
        throw new XMLCError(NOTATION_UNSUPPORTED);
    }

    /**
     * Generate ProcessingInstruction creation.
     */
    protected void genProcessInstructionCreate(ProcessingInstruction procInstr,
                                               String docVar,
                                               String varName, 
                                               JavaCode body) {
        body.addln(varName + " = " + docVar
                   + ".createProcessingInstruction("
                   + JavaLang.createStringConst(procInstr.getNodeName()) + ", "
                   + JavaLang.createStringConst(procInstr.getData()) + ");");
    }

    /**
     * Generate Text creation.
     */
    protected void genTextCreate(Text text,
                                 String docVar,
                                 String varName, 
                                 JavaCode body) {
        body.addln(varName + " = " + docVar + ".createTextNode("
                   + JavaLang.createStringConst(text.getData()) + ");");
    }

    /**
     * Generate code to construct a node, assigning it to the supplied
     * variable, Can be overriden by derived class
     */
    public void genNodeCreate(String docVar,
                              String varName,
                              Node node,
                              JavaCode body) throws XMLCException {
        switch (node.getNodeType()) {
        case Node.DOCUMENT_NODE:
        case Node.DOCUMENT_TYPE_NODE:
            throw new XMLCError("Creating node of this type must be done directly: "
                                + node.getClass().getName());

        case Node.ATTRIBUTE_NODE:
            genAttrCreate((Attr)node, docVar, varName, body);
            break;

        case Node.CDATA_SECTION_NODE:
            genCDATASectionCreate((CDATASection)node, docVar, varName,
                                  body);
            break;

        case Node.COMMENT_NODE:
            genCommentCreate((Comment)node, docVar, varName, body);
            break;

        case Node.ELEMENT_NODE:
            genElementCreate((Element)node, docVar, varName, body);
            break;

        case Node.ENTITY_NODE:
            genEntityCreate((Entity)node, docVar, varName, body);
            break;

        case Node.ENTITY_REFERENCE_NODE:
            genEntityReferenceCreate((EntityReference)node, docVar, varName, 
                                     body);
            break;

        case Node.NOTATION_NODE:
            genNotationCreate((Notation)node, docVar, varName, body);
            break;

        case Node.PROCESSING_INSTRUCTION_NODE:
            genProcessInstructionCreate((ProcessingInstruction)node,
                                        docVar, varName, body);
            break;

        case Node.TEXT_NODE:
            genTextCreate((Text)node, docVar, varName, body);
            break;

        case Node.DOCUMENT_FRAGMENT_NODE:
            throw new XMLCError("Creating node of this type not supported: "
                                + node.getClass().getName());

        default:
            throw new XMLCError("Unknown node type: " + node.getNodeType());
        }
    }

    /**
     * Determine if a node is a child of the document and needs to be inserted
     * before the document variable.
     */
    private boolean preceedsDocElement(Node node) {
        // Scan following siblings for document element
        for (Node sibling = node.getNextSibling(); sibling != null;
             sibling = sibling.getNextSibling()) {
            if (sibling == fDocElement) {
                    return true;  // doc element after
            }
        }
        return false;        
    }

    /**
     * Generate code to add a child of the document
     */
    private void genAppendDocumentChild(String parentVar,
                                        String childVar,
                                        Node child,
                                        JavaCode body) {
        // The Document element is automatically created, don't add it
        if (child != fDocElement) {
            // Make sure we add it on the correct side of the document element.
            if (preceedsDocElement(child)) {
                body.addln(parentVar + ".insertBefore(" + childVar + ", "
                           + parentVar + ".getDocumentElement());");
            } else {
                body.addln(parentVar + ".appendChild(" + childVar + ");");
            }
        }
    }

    /**
     * Generate code to do an append child.
     */
    public void genAppendChild(String parentVar,
                               String childVar,
                               Node child,
                               JavaCode body) {
        if (child.getParentNode() == fDocument) {
            // special handling for children of Document.
            genAppendDocumentChild(parentVar, childVar, child, body);
        } else {
            body.addln(parentVar + ".appendChild(" + childVar + ");");
        }
    }

    /**
     * Generate code to add attribute to an element.
     */
    public void genAddAttribute(String elementVar,
                                String attrVar,
                                JavaCode body) {
        body.addln(elementVar + ".setAttributeNode(" + attrVar + ");");
    }

    /**
     * Add an entity to a DocumentType
     */
    public void genAddEntity(String docTypeVar,
                             String entityVar,
                             Entity entity,
                             JavaCode body) {
        throw new XMLCError(ENTITIES_UNSUPPORTED);
    }

    /**
     * Add an notation to a DocumentType
     */
    public void genAddNotation(String docTypeVar,
                               String notationVar,
                               Notation notation,
                               JavaCode body) {
        throw new XMLCError(NOTATION_UNSUPPORTED);
    }
}
