/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: LazyDOMAccessorGenerator.java,v 1.2 2003/06/29 05:59:03 jkjome Exp $
 */

package org.enhydra.xml.xmlc.dom.lazydom;

import org.enhydra.xml.lazydom.*;
import org.enhydra.xml.xmlc.dom.AccessorGenerator;
import org.enhydra.xml.xmlc.dom.generic.GenericAccessorGenerator;
import org.enhydra.xml.xmlc.compiler.ElementInfo;
import org.enhydra.xml.xmlc.compiler.ClassGenerator;
import org.enhydra.xml.xmlc.codegen.JavaClass;
import org.enhydra.xml.xmlc.codegen.JavaMethod;
import org.enhydra.xml.xmlc.codegen.JavaField;
import org.enhydra.xml.xmlc.codegen.JavaCode;
import org.enhydra.xml.xmlc.codegen.JavaParameter;
import org.enhydra.xml.xmlc.codegen.JavaModifiers;
import org.w3c.dom.Document;

/*
 * FIXME:  Should syncAccessMethods just generate an error for the
 * LazyDOM?  This way, elementIds could be statics and access code
 * simpler.
 * Na, better yet, make syncAccessMethods a compiler-time option.
 */

/**
 * Generate access methods for the Lazy DOM.
 */
public class LazyDOMAccessorGenerator extends GenericAccessorGenerator
    implements AccessorGenerator {
    /**
     * Map of document nodes to node ids.
     */
    private NodeIdMap fNodeIdMap;

    /**
     * Constructor.
     */
    protected LazyDOMAccessorGenerator(Document document) {
        fNodeIdMap = new NodeIdMap(document);
    }

    /**
     * Get the name of the field that will store an elements' node id.
     */
    private String getElementNodeIdField(ElementInfo elementInfo) {
        return "$elementId_" + elementInfo.getXmlId();
    }

    /**
     * This method does nothing for the Lazy DOM.
     * @see AccessorGenerator#createAccessMethodInit
     */
    public void createAccessMethodInit(ElementInfo elementInfo,
                                       String elementVarName,
                                       JavaCode body) {
    }

    /**
     * Create code cached information so that element access methods
     * return null.
     * @see AccessorGenerator#createNullElementAccess
     */
    public void createNullElementAccess(ElementInfo elementInfo,
                                        JavaCode body) {
        body.addln(getElementNodeIdField(elementInfo) + " = -1;");
        body.addln(getElementField(elementInfo) + " = null;");
    }

    /**
     * Create code to reset the element access information from a node.
     * @see AccessorGenerator#createResetElementAccess
     */
    public void createResetElementAccess(ElementInfo elementInfo,
                                         String nodeVarName,
                                         JavaCode body) {
        body.addln(getElementNodeIdField(elementInfo) + " = "
                   + fNodeIdMap.getIdStr(elementInfo.getElement())
                   + ";");
        body.addln(getElementField(elementInfo) + " = ("
                   + elementInfo.getClassName() + ")" + nodeVarName + ";");
    }

    /**
     * Create code to test if an element is expanded and expand if necessary.
     */
    private void createExpandedElementCheck(ElementInfo elementInfo,
                                            JavaCode body) {
        String fieldName = getElementField(elementInfo);
        String idFieldName = getElementNodeIdField(elementInfo);
        body.addln(new String[] {
            "if ((" + fieldName + " == null) && (" + idFieldName + " >= 0)) {",
            "    " + fieldName + " = (" + elementInfo.getClassName() + ")"
            + LazyDOMDocBuilderGenerator.LAZY_DOCUMENT_FIELD + ".getNodeById(" + idFieldName + ");",
            "}",
        });
    }

    /**
     * Create an element accessor method (getElementXXX() or getTextXXX()).
     * @see AccessorGenerator#createAccessorMethod
     */
    public JavaMethod createAccessorMethod(ElementInfo elementInfo,
                                           ElementInfo.AccessorInfo accessor,
                                           boolean createDelegation,
                                           JavaClass docClass) {
        JavaMethod method = setupGetElementMethod(elementInfo, accessor,
                                                  createDelegation,
                                                  docClass);

        // Create return of value
        JavaCode body = method.getCode();
        createExpandedElementCheck(elementInfo, body);
        body.addln("return " + getElementField(elementInfo) +  ";");
        
        // Create fields to store element id and element.
        JavaField field = new JavaField(getElementNodeIdField(elementInfo),
                                        "int",
                                        JavaModifiers.PRIVATE,
                                        (String)null, 
                                        fNodeIdMap.getIdStr(elementInfo.getElement()));
        docClass.addField(field);
        field = new JavaField(getElementField(elementInfo),
                              elementInfo.getClassName(),
                              JavaModifiers.PRIVATE,
                              (String)null, null);
        docClass.addField(field);

        return method;
    }

    /**
     * Create the setText method for an element.
     * @see AccessorGenerator#createSetTextMethod
     */
    public JavaMethod createSetTextMethod(ElementInfo elementInfo,
                                          boolean createDelegation,
                                          JavaClass docClass) {
        JavaMethod method = setupSetTextMethod(elementInfo,
                                               createDelegation,
                                               docClass);

        // Create set of value
        JavaCode body = method.getCode();
        createExpandedElementCheck(elementInfo, body);
        body.addln("doSetText(" + getElementField(elementInfo) + ", text);");

        return method;
    }
}
