/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: PreFormatter.java,v 1.1.1.1 2003/03/10 16:36:22 taweili Exp $
 */

package org.enhydra.xml.xmlc.dom.lazydom;

import org.enhydra.xml.xmlc.dom.XMLCDocument;
import org.enhydra.xml.io.DOMFormatter;
import org.enhydra.xml.io.Formatter;
import org.enhydra.xml.io.OutputOptions;
import org.enhydra.xml.io.HTMLElements;
import org.enhydra.xml.xmlc.codegen.JavaCode;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Create pre-formatted text for inclusion in template DOM. Save's overhead of
 * formatting the text. This adds character entity references for
 * CharacterData nodes and formats the entire opening tag, with attributes,
 * for elements.  The default encoding is used to determine if this is an
 * 7, 8, or 16 bit encoding, and substituting character references as
 * needed.
 */
class PreFormatter {
    /** Formatter for this document. */
    private Formatter fFormatter;

    /** Is this an HTML document? */
    private boolean fIsHtmlDocument;

    /** Output options being used */
    private OutputOptions fOptions;

    /**
     * Constructor.
     */
    public PreFormatter(XMLCDocument xmlcDoc) {
        // Get default options and set encoding.  This ensure that the
        // encoding is available when the preformatting options are check at
        // runtime to determine if the preformatted text should be used.

        fOptions = DOMFormatter.getDefaultOutputOptions(xmlcDoc.getDocument());
        String encoding = xmlcDoc.getEncoding();
        if (encoding != null) {
            fOptions.setXmlEncoding(encoding);
        }

        fIsHtmlDocument = xmlcDoc.isHtmlDocument();
        fFormatter = DOMFormatter.getFormatter(xmlcDoc.getDocument(), fOptions, true);
    }

    /**
     * Check to see if this node is a descendent of a HTML script
     * or style element.  If so, no pre-formatting is done, as the
     * text is written as-is.
     */
    private boolean isHtmlFormatProtected(Node node) {
        if (!(node instanceof Element)) {
            // Find parent element.
            Node parent = node.getParentNode();
            while ((parent != null) && (!(parent instanceof Element))) {
                parent = parent.getParentNode();
            }
            if ((parent != null) && HTMLElements.isScriptStyle((Element)parent)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Get pre-formatted text for a node.
     * @return The text, of null if none is approriate for this node
     *  type.
     */
    public String preFormatNode(Node node) {
        if (fIsHtmlDocument && isHtmlFormatProtected(node)) {
            return null;
        } else {
            return fFormatter.preFormatNode(node);
        }
    }

    /**
     * Get the encoding being used for the preformatted text.
     */
    public String getMIMEEncoding() {
        return fFormatter.getMIMEEncoding();
    }

    /**
     * Generate code to recreate the output options used to format this code.
     */
    public void createOutputOptionsCodeGenerator(String varName,
                                                 JavaCode code) {
        fOptions.createCodeGenerator(varName, true, code);
    }
}
