/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HTMLProprietaryTags.java,v 1.1.1.1 2003/03/10 16:36:22 taweili Exp $
 */

package org.enhydra.xml.xmlc.html;

import org.enhydra.xml.xmlc.*;
import java.util.Hashtable;
import java.util.Enumeration;

//FIXME: is this still used???

/**
 * Table of proprietary tags and attributes for the HTML parser.
 */
public class HTMLProprietaryTags {
    /**
     * Abobe CyberStudio proprietary tag set name.
     */
    public static final String TAG_SET_CYBER_STUDIO = "cyberstudio";

    /**
     * Empty tag content model.
     */
    public static final int TAG_CM_EMPTY = 0x01;

    /**
     * Inline tag content model.
     */
    public static final int TAG_CM_INLINE = 0x02;

    /**
     * Block tag content model.
     */
    public static final int TAG_CM_BLOCK = 0x04;

    /**
     * Closing tag is optional.
     */
    public static final int TAG_CM_OPT = 0x08;

    /**
     * Class used to record an tag with its content model.
     */
    public class Tag {
        /**
         * Upper-case tag name.
         */
        public final String tagName;

        /**
         * Content model the tag is associated with. (TAG_CM_*).
         */
        public final int contentModel;
        
        /**
         * The tag set this tag is associated with, or null if it was
         * not added as part of a tag set.
         */
        public final String tagSet;

        /**
         * Create a new tag.
         */
        protected Tag(String tagName,
                      int contentModel) {
            this.tagName = tagName;
            this.contentModel = contentModel;
            this.tagSet = null;
        }

        /**
         * Create a new tag.
         */
        protected Tag(String tagName,
                      int contentModel,
                      String tagSet) {
            this.tagName = tagName;
            this.contentModel = contentModel;
            this.tagSet = tagSet;
        }
    }

    /**
     * Hash table of tags that have been specified to their content model.
     */
    private Hashtable tags = new Hashtable();

    /**
     * Hash table of attributes that have been specified.  Hash is used
     * to prevent duplicates.
     */
    private Hashtable attributes = new Hashtable();
    
    /**
     * Hash table of tag sets that have been defined.
     */
    private Hashtable tagSets = new Hashtable();

    /**
     * Specify one of the pre-defined sets of proprietary tags that the parser
     * should accept.  If the parser is able to handle arbitrary tags, this
     * should be ignored.  If the parser can only parse valid HTML, an error
     * should be generated.
     *
     * @param tagSetName One of the predefined tag set names.  An error is
     * generated if its illegal.
     */
    public void addTagSet(String tagSetName) throws XMLCException {
        if (tagSetName.equalsIgnoreCase(TAG_SET_CYBER_STUDIO)) {
            if (!tagSets.containsKey(TAG_SET_CYBER_STUDIO)) {
                registerCyberStudioTags();
                tagSets.put(TAG_SET_CYBER_STUDIO, TAG_SET_CYBER_STUDIO);
            }
        } else {
            throw new XMLCException ("Unknown proprietary tag set: \""
                                     + tagSetName + "\", valid sets are: "
                                     + TAG_SET_CYBER_STUDIO);
        }
    }

    /**
     * Get an enumeration of the tag sets that have been defined.
     */
    public Enumeration getTagSets() {
        return tagSets.elements();
    }

    /**
     * Add a proprietary tag to the set of allowed tags.
     * 
     * @param tagName The name of the tag.  The name is case insensitive.
     * @param contentModel A bit set of the <CODE>TAG_CM_*</CODE> constants.
     */
    public void addTag(String tagName,
                       int contentModel) throws XMLCException {
        tagName = tagName.toUpperCase();
        tags.put(tagName, new Tag(tagName,
                                  contentModel));
    }
    
    /**
     * Get an enumeration of the proprietary Tag objects.
     */
    public Enumeration getTags() {
        return tags.elements();
    }

    /**
     * Add a proprietary attribute to the set of allowed attributes.
     * 
     * @param attributeName The attribute name. The name is case insensitive.
     */
    public void addAttribute(String attributeName) throws XMLCException {
        attributeName = attributeName.toUpperCase();
        attributes.put(attributeName, attributeName);
    }

    /**
     * Get an enumeration of the Tag objects for the proprietary tags.
     */
    public Enumeration getAttributes() {
        return attributes.elements();
    }

    /**
     * Add a proprietary tag from a tagset.
     */
    public void addTag(String tagName,
                       int contentModel,
                       String tagSet) {
        tagName = tagName.toUpperCase();
        tags.put(tagName, new Tag(tagName, contentModel, tagSet));
    }
    
    /**
     * Register Abobe CyberStudioTags.
     */
    private void registerCyberStudioTags() throws XMLCException {
        String[] csBlockTags = {
            "csactionclass", "csactiondict", "csactionitem",
            "csactionparam", "csactions",    "csgroup",
            "csimport",      "csmagic",      "cspart",
            "csscriptdict",  "cssequence",   "cssequencer",
            "cstrack"
        };
        for (int i = 0; i < csBlockTags.length; i++) {
            addTag(csBlockTags[i], TAG_CM_BLOCK, TAG_SET_CYBER_STUDIO);
        }

        addTag("csobj", TAG_CM_BLOCK|TAG_CM_INLINE, TAG_SET_CYBER_STUDIO);

        String[] csEmptyTags = {
            "csbrowser",    "csaction",  "csactioncntl",
            "csactionenum", "csgeneric", "csseqaction"
        };
        for (int i = 0; i < csEmptyTags.length; i++) {
            addTag(csEmptyTags[i], TAG_CM_INLINE|TAG_CM_EMPTY|TAG_CM_OPT,
                   TAG_SET_CYBER_STUDIO);
        }
    }
}
