/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ParserAdaptor.java,v 1.1.1.1 2003/03/10 16:36:22 taweili Exp $
 */

package org.enhydra.xml.xmlc.html.parsers.swing;

import org.enhydra.xml.xmlc.XMLCError;
import java.io.Reader;
import java.io.IOException;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.html.HTMLEditorKit;
import javax.swing.text.ChangedCharSetException;
import javax.swing.text.html.parser.ParserDelegator;
import javax.swing.text.html.parser.DocumentParser;
import javax.swing.text.html.parser.DTD;
import javax.swing.text.html.parser.Element;
import javax.swing.text.html.parser.TagElement;

/**
 * Wrapper around the Swing HTML parser that attempts to correct for
 * various problems in different versions of the Swing.  This code
 * is tightly bound to the Swing code. 
 * <P>
 * This corrects the following problems
 * <UL>
 * <LI> Swing 1.1:
 *   <UL>
 *    <LI>LINK elements were not allowed in the HEAD.
 *   </UL>
 * </UL>
 */
class ParserAdaptor {
    /** Swing version constants, can be compared numerically */
    public static final int SWING_1_1 = 11;
    public static final int SWING_1_2 = 12;
    public static final int SWING_1_3 = 13;

    /** Version of Swing being used */
    private static int swingVersion;

    /**
     * Magic tag name that is generate as a simple tag at the end of the parse
     * in Swing 1.2.  This is a little string, don't know why they did this
     * (or why is called `end of line'.
     */
    public static final String MAGIC_END_TAG = "__EndOfLineTag__";

    /**
     * Name of attribute added to HEAD tag when an empty one is created
     * by the parser.
     */
    public static final String IMPLIED_PSEUDO_ATTR = "_implied_";

    /**
     * Class initializer.
     */
    static {
        // Try to determine Swing version
        String specVersion = System.getProperty("java.specification.version");
        if ((specVersion == null) || specVersion.startsWith("1.1")) {
            swingVersion = SWING_1_1;
        } else if (specVersion.startsWith("1.2")) {
            swingVersion = SWING_1_2;
        } else {
            // Assume 1.3 or better
            swingVersion = SWING_1_3;
        }
    }

    /*
     * Class based on ParserDelegator that gets the DTD object.
     */
    private static class DTDCreator extends ParserDelegator {
        /** DTD name used by swing (through at least 1.3) */
        private static String DTD_NAME = "html32";
        
        /** The DTD we have obtained */
        private static DTD dtd = null;

        /**
         * Build the DTD on first access.
         */
        private static synchronized void buildDTD() throws IOException {
            DTD newDTD = DTD.getDTD(DTD_NAME);
            newDTD = createDTD(newDTD, DTD_NAME);

            // Set once initialized.
            dtd = newDTD;
        }

        /**
         * The DTD.
         */
        public static DTD getDTD() throws IOException {
            if (dtd == null) {
                buildDTD();
            }
            return dtd;
        }
    }

    /**
     * Base for all DocumentParser adaptors.
     */
    class DocumentParserBase extends DocumentParser {
        /** saved callback pointer */
        protected HTMLEditorKit.ParserCallback fCallback;
        
        /**
         * Constructor.
         */
        public DocumentParserBase(DTD dtd) {
            super(dtd);
        }

        /**
         * Override parse to get callback pointer.
         */
        public void parse(Reader in,
                          HTMLEditorKit.ParserCallback callback,
                          boolean ignoreCharSet) throws IOException {
            fCallback = callback;
            super.parse(in, callback, ignoreCharSet);
        }
    }

    /**
     * DocumentParser for 1.1 that allows LINKs in headers.
     */
    class DocumentParser11 extends DocumentParserBase {
        /**
         * Constructor.
         */
        public DocumentParser11(DTD dtd) {
            super(dtd);
        }
        
        /**
         * override handleEmptyTag to allow links in headers.
         */
        protected void handleEmptyTag(TagElement tag) throws ChangedCharSetException {
            Element elem = tag.getElement();
            if (elem.name.equalsIgnoreCase("link")) {
                if (tag.fictional()) {
                    fCallback.handleSimpleTag(tag.getHTMLTag(), new SimpleAttributeSet(), getCurrentPos());
                } else {
                    fCallback.handleSimpleTag(tag.getHTMLTag(), getAttributes(), getCurrentPos());
                    flushAttributes();
                }
            } else {
                super.handleEmptyTag(tag);
            }
        }
    }

    /**
     * Get the guessed version of swing being used.
     */
    public int getSwingVersion() {
        return swingVersion;
    }

    /**
     * Create a parser base on the version of swing we are linked to.
     */
    public DocumentParser getParser() {
        try {
            DTD dtd = DTDCreator.getDTD();
            if (swingVersion == SWING_1_1) {
                return new DocumentParser11(dtd);
            } else if (swingVersion == SWING_1_2) {
                return new DocumentParser(dtd);
            } else {
                return new DocumentParser(dtd);
            }
        } catch (Exception except) {
            throw new XMLCError("Can't created Swing HTML parser", except);
        }
    }
}
