/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: LineNumberRecorder.java,v 1.1.1.1 2003/03/10 16:36:23 taweili Exp $
 */

package org.enhydra.xml.xmlc.misc;

import java.util.EmptyStackException;
import java.util.ArrayList;

/**
 * Class that records information to translate a line number or
 * byte offset in a stream into a source file and line number.
 */
public final class LineNumberRecorder extends LineNumberMap {
    /*
     * Structure used to implement stack of file names and line numbers.
     */
    private class FileStackEntry {
        /** Saved file name, line number, and char offset */
        public final String fileName;
        public final int lineNum;

        /** Constructor */
        public FileStackEntry(String fileName,
                              int lineNum) {
            this.fileName = fileName;
            this.lineNum = lineNum;
        }
    }

    /**
     * Current source file name.
     */
    private String fFileName;

    /**
     * Current line-number in the file.
     */
    private int fLineNum = 1;
    
    /**
     * Line number in the stream.
     */
    private int fStreamLineNum = 1;

    /**
     * Character offset in the stream.
     */
    private int fStreamCharOffset;

    /**
     * Was the previous character a return? 
     */
    private boolean fPrevWasReturn;

    /**
     * Stack of FileStackEntry objects
     */
    private ArrayList fFileStack;

    /**
     * Constructor, setting initial file.
     */
    public LineNumberRecorder(String fileName) {
        pushFile(fileName);
    }

    /**
     * Constructor with no initial file.
     */
    public LineNumberRecorder() {
    }

    /**
     * Push a file on the stack.  Called when an included file is opened.
     */
    public void pushFile(String fileName) {
        // Don't push an entry if this is the initial file.
        if (fFileName != null) {
            if (fFileStack == null) {
                fFileStack = new ArrayList();
            }
            fFileStack.add(new FileStackEntry(fFileName, fLineNum));
        }
        fFileName = fileName;
        fLineNum = 1;
        // Add a line with a zero offset
        addLine();
    }

    /**
     * Pop a file from the stack.  Called when an included file is closed.
     * Allows for an extra pop to close the originally opened file.
     */
    public void popFile() {
        if ((fFileStack == null) || (fFileStack.size() == 0)) {
            if (fFileName == null) {
                throw new EmptyStackException();
            }
            // Pop original file.
            fFileName = null;
        } else {
            FileStackEntry entry = (FileStackEntry)fFileStack.get(fFileStack.size()-1);
            fFileName = entry.fileName;
            fLineNum = entry.lineNum;
            fFileStack.remove(fFileStack.size()-1);
        }
    }    

    /**
     * Add a line to the map.
     */
    private void addLine() {
        addLine(fFileName, fLineNum, fStreamLineNum, fStreamCharOffset);
    }

    /**
     * Count a character read (and hope this gets inlined).
     */
    public final void countChar(char ch) {
        fStreamCharOffset++;
        if (fPrevWasReturn) {
            fPrevWasReturn = false;
        } else {
            switch (ch) {
            case '\r':
                fPrevWasReturn = true;
            case '\n':
                fLineNum++;
                fStreamLineNum++;
                addLine();
            }
        }
    }

    /**
     * Count an array character's readed
     */
    public final void countChars(final char[] chBuf,
                                 int off,
                                 int len) {
        int endOff = off+len-1;
        for (int idx = 0; idx <= endOff; idx++) {
            countChar(chBuf[idx]);
        }
    }
}
