/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: BasicTests.java,v 1.1.1.1 2003/03/10 16:37:12 taweili Exp $
 */

package org.enhydra.xml.xmlc.servlet.src;

import org.enhydra.xml.io.OutputOptions;
import org.enhydra.xml.xmlc.servlet.XMLCContext;
import org.enhydra.xml.xmlc.XMLCFactory;
import org.enhydra.xml.xmlc.XMLObject;
import javax.servlet.ServletException;
import javax.servlet.ServletContext;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import java.io.IOException;
import java.io.FileNotFoundException;
import java.util.Date;
import java.text.DateFormat;
import java.io.File;
import java.io.RandomAccessFile;
import java.net.URL;


/**
 * Basic test of XMLCContext.  Test's creation and output of documents.
 * Performs several tasks based on parameters:
 * <UL>
 * <LI> doc - Parameter control what document to use:
 *    <UL>
 *    <LI> html1name - HTML page #1 by created by class name.
 *    <LI> html1class - HTML page #1 by created by interface class (default).
 *    <LI> wml1name - WML page #1 by created by class name.
 *    <LI> wml1class - WML page #1 by created by interface class.
 *    <LI> big5name - Big5 HTML page #1 by name.
 *    </UL>
 * <LI> touch - Parameter used to control touching page/class.
 *    <UL>
 *    <LI> none - Don't touch files (default).
 *    <LI> class - Touch class file.
 *    <LI> source - Touch source file.
 *    </UL>
 * <LI> encoding - Use encoding.
 * </UL>
 */
public class BasicTests extends HttpServlet {
    /**
     * Path names.
     */
    private static final String PKG_NAME = "tests/org/enhydra/xml/xmlc/servlet/src";
    private static final String HTML1_CLASS = PKG_NAME + "/TestPage1HTML.class";
    private static final String HTML1_SOURCE = PKG_NAME + "/TestPage1.html";
    private static final String WML1_CLASS = PKG_NAME + "/WMLPage1WML.class";
    private static final String WML1_SOURCE = PKG_NAME + "/TestPage1.wml";
    private static final String BIG5_CLASS = PKG_NAME + "/Big5HTML.class";
    private static final String BIG5_SOURCE = PKG_NAME + "/big5.page1.html";

    /**
     * doc parameter name and values.
     */
    private static final String DOC_PARAM = "doc";
    private static final String DOC_HTML1NAME = "html1name";
    private static final String DOC_HTML1CLASS = "html1class";
    private static final String DOC_WML1NAME = "wml1name";
    private static final String DOC_WML1CLASS = "wml1class";
    private static final String DOC_BIG5NAME = "big5name";

    /**
     * touch parameter name and values.
     */
    private static final String TOUCH_PARAM = "touch";
    private static final String TOUCH_NONE = "none";
    private static final String TOUCH_CLASS = "class";
    private static final String TOUCH_SOURCE = "source";

    /**
     * encoding parameter.
     */
    private static final String ENCODING_PARAM = "encoding";

    // Number of bytes to write in file.
    private static final int FILE_WRITE_SIZE = 128;

    // Updates files in the documents
    private TestInfoUpdate testInfoUpdate = new TestInfoUpdate();

    /**
     * Parse the doc parameter, returning one of the constants
     * (so that refs can be compared).
     */
    private String parseDocParam(HttpServletRequest request) {
        String docParam = request.getParameter(DOC_PARAM);
        if (docParam == null) {
            return DOC_HTML1CLASS;
        } else if (docParam.equals(DOC_HTML1NAME)) {
            return DOC_HTML1NAME;
        } else if (docParam.equals(DOC_HTML1CLASS)) {
            return DOC_HTML1CLASS;
        } else if (docParam.equals(DOC_WML1NAME)) {
            return DOC_WML1NAME;
        } else if (docParam.equals(DOC_WML1CLASS)) {
            return DOC_WML1CLASS;
        } else if (docParam.equals(DOC_BIG5NAME)) {
            return DOC_BIG5NAME;
        } else {
            throw new XMLTestException("unknown value for \"doc\" parameter: \""
                                       + docParam + "\"");
        }
    }

    /**
     * Parse touch param and return one of the constants
     * (so that refs can be compared).
     */
    private String parseTouchParam(HttpServletRequest request) {
        String touchParam = request.getParameter(TOUCH_PARAM);
        if (touchParam == null) {
            return TOUCH_NONE;
        } else if (touchParam.equals(TOUCH_NONE)) {
            return TOUCH_NONE;
        } else if (touchParam.equals(TOUCH_CLASS)) {
            return TOUCH_CLASS;
        } else if (touchParam.equals(TOUCH_SOURCE)) {
            return TOUCH_SOURCE;
        } else {
            throw new XMLTestException("unknown value for \"touch\" parameter: \""
                                       + touchParam + "\"");
        }
    }

    /**
     * Locate the class or source file on the class path.
     * <P>
     * FIXME: Note: Tomcat appears to get confused about the location, leaving
     * out the WEB-INF/classes part of the path.  We work around this by
     * checking if the first file is not found.
     */
    private File getPackageFile(String path) throws IOException {
        ServletContext servletContext = getServletConfig().getServletContext();
        URL url = servletContext.getResource(path);
        if (url == null) {
            throw new IOException("can't find resource: \"" + path + "\"");
        }
        if (!url.getProtocol().equals("file")) {
            throw new IOException("protocol for resource not `file:': \"" + url + "\"");
        }
        File file = new File(url.getFile());

        if (file.exists()) {
            return file;
        } else {
            // Try inserting WEB-INF/classes
            String absPath = file.getAbsolutePath();
            String baseDir = absPath.substring(0, absPath.length()-path.length());
            File file2 = new File(baseDir + "/WEB-INF/classes/" + path);
            if (!file2.exists()) {
                String msg = "can't find resource \"" + path
                    + "\" on class path; tried " + "\"" + file + "\""
                    + " and \"" + file2 + "\"";
                throw new FileNotFoundException(msg);
            }
            return file2;
        }
    }

    /**
     * Create document based on doc parameter.
     */
    private XMLObject createDoc(XMLCContext context,
                                HttpServletRequest request) {
        String docParam = parseDocParam(request);
        XMLCFactory factory = context.getXMLCFactory();
        if (docParam == DOC_HTML1NAME) {
            return factory.create(TestPage1HTML.class.getName());
        } else if (docParam == DOC_HTML1CLASS) {
            return factory.create(TestPage1HTML.class);
        } else if (docParam == DOC_WML1NAME) {
            return factory.create(WMLPage1WML.class.getName());
        } else if (docParam == DOC_WML1CLASS) {
            return factory.create(WMLPage1WML.class);
        } else if (docParam == DOC_BIG5NAME) {
            return factory.create(Big5HTML.class.getName());
        } else {
            return null;  // Can't happen
        }
    }

    /**
     * Get the class or source to modify.
     */
    private File getModifyFile(HttpServletRequest request,
                               boolean getSource) throws IOException {
        String docParam = parseDocParam(request);
        if ((docParam == DOC_HTML1NAME) || (docParam == DOC_HTML1CLASS)) {
            if (getSource) {
                return getPackageFile(HTML1_SOURCE);
            } else {
                return getPackageFile(HTML1_CLASS);
            }
        } else if ((docParam == DOC_WML1NAME)
                    || (docParam == DOC_WML1CLASS)) {
            if (getSource) {
                return getPackageFile(WML1_SOURCE);
            } else {
                return getPackageFile(WML1_CLASS);
            }
        } else if ((docParam == DOC_BIG5NAME)) {
            if (getSource) {
                return getPackageFile(BIG5_SOURCE);
            } else {
                return getPackageFile(BIG5_CLASS);
            }
        } else {
            return null; // Can't happen
        }
    }

    /**
     * Modify the class or source file.
     */
    private void modifyFile(HttpServletRequest request,
                            boolean modSource) throws IOException {
        File file = getModifyFile(request, modSource);
        RandomAccessFile raf = new RandomAccessFile(file, "rw");
        byte[] buf = new byte[FILE_WRITE_SIZE];
        int readCnt = raf.read(buf);
        raf.seek(0);
        raf.write(buf, 0, readCnt);
    }

    /**
     * Check handle modification of files for reloading.
     */
    private void modifyIfRequested(HttpServletRequest request) throws IOException {
        String touchParam = parseTouchParam(request);
        if (touchParam == TOUCH_CLASS) {
            modifyFile(request, false);
        } else if (touchParam == TOUCH_SOURCE) {
            modifyFile(request, true);
        }
    }

    /**
     * Output page.
     */
    private void writePage(XMLObject doc,
                           XMLCContext context,
                           HttpServletRequest request,
                           HttpServletResponse response,
                           String encoding) throws ServletException, IOException {
        if (encoding != null) {
            OutputOptions options
                = context.createOutputOptions(request, response, doc);
            if (encoding != null) {
                options.setEncoding(encoding);
            }
            context.writeDOM(request, response, options, doc);
        } else {
            context.writeDOM(request, response, doc);
        }
    }

    /**
     * Entry.
     */
    public void doGet(HttpServletRequest request,
                      HttpServletResponse response) throws ServletException, IOException {

        HttpSession session = request.getSession(true);
        modifyIfRequested(request);

        XMLCContext context = XMLCContext.getContext(this); 
        XMLObject doc = createDoc(context, request);
        testInfoUpdate.setTestInfo(doc, context, request);
        writePage(doc, context, request, response, 
                  request.getParameter(ENCODING_PARAM));
    }
}
